"""
File Purpose: Faceplot: 3 plots on 3 faces of a box in 3D.

[TODO] faceplot with variable plot dim, via PlotDimsMixin.
"""

import matplotlib.pyplot as plt
import mpl_toolkits.mplot3d as mpl3d
import numpy as np
import xarray as xr

from .labels import xarray_title_plot_node
from .movies import MoviePlotElement, MoviePlotNode
from .plot_settings import PlotSettings
from .plot_tools import (
    current_axes_or_None,
    update_vlims,
    infer_movie_dim,
)

from ..errors import PlottingAmbiguityError, PlottingNotImplementedError
from ..tools import (
    simple_property, alias, alias_child,
    # format_docstring, UNSET,
    is_iterable,
    xarray_isel, xarray_nondim_coords,
    pcAccessor,
)
from ..defaults import DEFAULTS


''' --------------------- Helpful tools --------------------- '''

@pcAccessor.register('faces_3D', totype='array')
def xarray_faces_3D(array, ikeep=0, *, dims=('x', 'y', 'z')):
    '''infer the faces (at 0th index of dims) of a 3D xarray.DataArray.
    ikeep: int
        select ikeep'th value in the removed-dimension for each face.
        E.g. if ikeep=0: x_y=array.isel(z=0), x_z=array.isel(y=0), and y_z=array.isel(x=0).
    dims: 3-tuple of str
        the names of the dimensions for the x, y, and z faces.
        also affects resulting data var names,
            e.g. dims='abc' --> result will have 'a_b' instead of 'x_y'.
    
    returns xarray.Dataset with data vars x_y, x_z, y_z.
    '''
    x, y, z = dims
    result = {
        'x_y': array.isel({z: ikeep}),
        'x_z': array.isel({y: ikeep}),
        'y_z': array.isel({x: ikeep}),
    }
    return xr.Dataset(result)


''' --------------------- Faceplot stuff --------------------- '''

@PlotSettings.format_docstring(sub_ntab=1)
class FaceplotPlotElement(MoviePlotElement):
    '''3 plots on 3 faces of a box in 3D.
    Faceplot.__init__ will make these plots.

    Troubleshooting colors? Try providing vmin, vmax, & levels, explicitly!
    
    coords: dictlike with keys 'x', 'y', 'z'
        1D coordinate arrays along each dimension.
    data: dictlike with keys 'x_y', 'x_z', 'y_z'
        2D arrays of data to plot on the faces.
        E.g. 'x_z' data will go on the x-z face.
    ax: None or axes with 3d projection.
        e.g. fig.add_subplot(1,1,1, projection='3d')
        if None, use plt.gca() if current axes exist, else make new axes.

    Additional kwargs control settings for the plot:
        faceplot_view_angle: {faceplot_view_angle}
        faceplot_edge_kwargs: {faceplot_edge_kwargs}
        faceplot_axes_zoom: {faceplot_axes_zoom}
        aspect3d: {aspect3d}

        any of these kwargs for ax.contourf:
            'vmin', 'vmax', 'levels', 'cmap',

        any of these kwargs for ax.set:
            'xlabel', 'ylabel', 'zlabel', 'xlim', 'ylim', 'zlim',
            'xticks', 'yticks', 'zticks', 'xticklabels', 'yticklabels', 'zticklabels'
            Note the defaults for 'xlabel', 'ylabel', 'zlabel' will be 'x', 'y', 'z'.

        add_colorbar: bool
            whether to self.colorbar() during __init__.
        colorbar_kw: {colorbar_kw}

    These attrs of self will be created/updated during init (here, face='x_y', 'x_z', or 'y_z'):
        meshgrids: {{face: {{x: 2d array of x values at face}} for x in face}}
        coord_lims: {{x: [min, max] for x in 'x', 'y', 'z'}}
        data_lims: {{'all': [min, max], **{{face: [min, max]}} }}
        ax: the axes object to plot on; create if needed.
        faces: {{face: the mpl_toolkits.mplot3d.art3d.QuadContourSet3D plotted on this face}}
        edges: {{x: the mpl_toolkits.mplot3d.art3d.Line3D plotted at x=0. x='x', 'y' or 'z'}}
    '''
    # # # CREATING & INITIALIZING # # #
    def __init__(self, coords, data, *, ax=None,
                 xlabel='x', ylabel='y', zlabel='z',
                 add_colorbar=True,
                 **kw_super):
        self.coords = coords
        self._data_init = data  # <-- may help with debugging
        self._ax_init = ax
        kw_super.update(xlabel=xlabel, ylabel=ylabel, zlabel=zlabel, add_colorbar=add_colorbar)
        super().__init__(**kw_super)
        # super().__init__ assigns self.data, so this line must go after that:
        self.data = {face: data[face] for face in ('x_y', 'x_z', 'y_z')}
        # create the whole plot:
        self.init_all()

    fig = alias_child('ax', 'figure', if_no_child=None, doc='''figure containing this faceplot.''')

    def init_all(self):
        '''call all relevant init_* methods.'''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        self.init_meshgrids()
        self.init_coord_lims()
        self.init_data_lims()
        self.init_ax()
        self.init_faces()
        self.init_edges()
        self.init_viewing()
        if self.plot_settings['add_colorbar']:
            self.colorbar()

    @classmethod
    def from_dataset(cls, ds, **kw):
        '''create Faceplot from an xarray.Dataset.
        Equivalent: cls(ds.coords, ds, **kw)
        '''
        return cls(ds.coords, ds, **kw)

    def init_meshgrids(self):
        '''initialize self.meshgrids based on coords from self.coords.
        self.meshgrids = {
            'x_y': {'x': 2d array of x coords in x_y plot,
                    'y': 2d array of y coords in x_y plot},
            'x_z': {'x': 2d array of x coords in x_z plot,
                    'z': 2d array of z coords in x_z plot},
            'y_z': {'y': 2d array of y coords in y_z plot,
                    'z': 2d array of z coords in y_z plot},
        }
        returns self.meshgrids.
        '''
        coords = self.coords
        grid_lists = dict()
        grid_lists['x_y'] = np.meshgrid(coords['x'], coords['y'], indexing='ij')
        grid_lists['x_z'] = np.meshgrid(coords['x'], coords['z'], indexing='ij')
        grid_lists['y_z'] = np.meshgrid(coords['y'], coords['z'], indexing='ij')
        result = dict()
        result['x_y'] = {'x': grid_lists['x_y'][0], 'y': grid_lists['x_y'][1]}
        result['x_z'] = {'x': grid_lists['x_z'][0], 'z': grid_lists['x_z'][1]}
        result['y_z'] = {'y': grid_lists['y_z'][0], 'z': grid_lists['y_z'][1]}
        self.meshgrids = result
        return self.meshgrids

    def init_coord_lims(self):
        '''initialize self.coord_lims based on self.coords.
        self.coord_lims = {x: [min, max] for x in 'x', 'y', 'z'}
        returns self.coord_lims.
        '''
        coords = self.coords
        self.coord_lims = {x: [coords[x].min(), coords[x].max()] for x in ('x', 'y', 'z')}
        return self.coord_lims

    def init_data_lims(self):
        '''initialize self.data_lims based on self.data.
        data_lims = {
            'x_y': [min, max] of data['x_y'],
            'x_z': [min, max] of data['x_z'],
            'y_z': [min, max] of data['y_z'],
            'all': [min, max] of all data,
        }
        returns self.data_lims.
        '''
        data = self.data
        data_lims = dict()
        data_lims['x_y'] = [data['x_y'].min(), data['x_y'].max()]
        data_lims['x_z'] = [data['x_z'].min(), data['x_z'].max()]
        data_lims['y_z'] = [data['y_z'].min(), data['y_z'].max()]
        all_min = min(data_lims['x_y'][0], data_lims['x_z'][0], data_lims['y_z'][0])
        all_max = max(data_lims['x_y'][1], data_lims['x_z'][1], data_lims['y_z'][1])
        data_lims['all'] = [all_min, all_max]
        self.data_lims = data_lims

    def init_ax(self):
        '''initialize ax for self. returns self.ax'''
        ax = self._ax_init
        if ax is None:
            ax = current_axes_or_None()
        if ax is None:
            fig = plt.gcf()
            kw_ax = self.plot_settings.get_mpl_kwargs('mpl_toolkits.mplot3d.axes3d.Axes3D')
            ax = fig.add_subplot(1,1,1, projection='3d', **kw_ax)
        else:
            if not isinstance(ax, mpl3d.axes3d.Axes3D):
                errmsg = f'expected instance of mpl_toolkits.mplot3d.axes3d.Axes3D, got ax of type {type(ax)}'
                raise TypeError(errmsg)
        self.ax = ax
        return self.ax

    def init_faces(self):
        '''initialize the faces on self.ax; actually plot the data.
        stores plotted objects in self.faces & returns self.faces.
        '''
        ax = self.ax
        data = self.data
        grids = self.meshgrids
        kw_plot = self.plot_settings.get_mpl_kwargs('plt.contourf')
        # determine vlims for each face.
        kw_plot = update_vlims(list(data.values()), kw_plot)
        self.kw_plot = kw_plot  # remember kw for self._plot_face later.
        # actually plot the faces:
        self.faces = dict()
        for face in ('x_y', 'x_z', 'y_z'):
            self._plot_face(face, data[face])
        return self.faces

    def _plot_face(self, face, data2d):
        '''plots this data (2d array) on this face.
        updates self.faces accordingly. removes old face if it exists.
        returns the newly plotted face.
        '''
        ax = self.ax
        grids = self.meshgrids
        if face in self.faces:
            to_remove = self.faces[face]
        else:
            to_remove = None
        kw_plot = self.kw_plot
        if face == 'x_y':
            result = ax.contourf(grids['x_y']['x'], grids['x_y']['y'], data2d, zdir='z', offset=0, **kw_plot)
        elif face == 'x_z':
            result = ax.contourf(grids['x_z']['x'], data2d, grids['x_z']['z'], zdir='y', offset=0, **kw_plot)
        elif face == 'y_z':
            result = ax.contourf(data2d, grids['y_z']['y'], grids['y_z']['z'], zdir='x', offset=0, **kw_plot)
        else:
            raise ValueError(f'got face={face!r}; expected "x_y", "x_z", or "y_z".')
        self.faces[face] = result
        # remove old face:
        if to_remove is not None:
            to_remove.remove()  # remove old face from the figure, first.
        # return plotted face
        return result

    def init_edges(self):
        '''initialize the edges on self.ax; actually plot the edges.
        stores plotted objects in self.edges & returns self.edges.

        edges will not be plotted if self.plot_settings.get('faceplot_edge_kwargs') is None.
        '''
        kw_edges = self.plot_settings.get('faceplot_edge_kwargs')
        if kw_edges is None:
            self.edges = None
            return self.edges
        ax = self.ax
        lims = self.coord_lims
        edges = dict()
        edges['x'] = ax.plot(lims['x'], [lims['y'][0], lims['y'][0]], [lims['z'][0], lims['z'][0]], **kw_edges)
        edges['y'] = ax.plot([lims['x'][0], lims['x'][0]], lims['y'], [lims['z'][0], lims['z'][0]], **kw_edges)
        edges['z'] = ax.plot([lims['x'][0], lims['x'][0]], [lims['y'][0], lims['y'][0]], lims['z'], **kw_edges)
        self.edges = edges
        return self.edges

    def init_viewing(self):
        '''initialize view-related things: labels, viewing angle, aspect ratio, etc.'''
        ax = self.ax
        # set viewing angle
        view_angle = self.plot_settings.get('faceplot_view_angle')
        if view_angle is not None:  # (None --> use matplotlib default instead of setting angle.)
            ax.view_init(*view_angle)
        # set aspect ratio
        aspect = self.plot_settings.get('aspect3d')
        if aspect == 'auto':
            sizes = None
        elif aspect == 'equal' or aspect == 1 or (is_iterable(aspect) and len(aspect)==4):
            # matplotlib doesn't make it easy... need to calculate manually.
            coords = self.coords
            sizes = [len(coords[x]) for x in 'xyz']
            if is_iterable(aspect) and len(aspect)==4:
                sizes = [s*mul for s, mul in zip(sizes, aspect[1:])]
        elif is_iterable(aspect) and len(aspect)==3:
            sizes = aspect
        else:
            errmsg = f'aspect={aspect} not yet implemented; expected "auto", "equal", 3-tuple, or 4-tuple.'
            raise PlottingNotImplementedError(errmsg)
        if sizes is not None:
            zoom = self.plot_settings.get('faceplot_axes_zoom')
            ax.set_box_aspect(sizes, zoom=zoom)
        # set labels
        ax_setting = self.plot_settings.get_mpl_kwargs('ax.set')
        ax.set(**ax_setting)

    # # # BEHAVIORS # # #
    def __getitem__(self, k):
        '''returns self.faces[k] for 'x_y', 'x_z', or 'y_z', else self.edges[k] for 'x', 'y', or 'z',
        else crash with KeyError.
        '''
        if k in ('x_y', 'x_z', 'y_z'):
            return self.faces[k]
        elif k in ('x', 'y', 'z'):
            return self.edges[k]
        else:
            raise KeyError(f'key {k} not in {self.__class__.__name__}')

    # # # UPDATING # # #
    def update_data(self, data):
        '''updates the data on the plot to match the data provided here.
        data: dictlike with keys 'x_y', 'x_z', 'y_z'
            2D arrays of data to plot on the faces.
            E.g. 'x_z' data will go on the x-z face.
        returns the list of updated face artist objects.
        '''
        result = []
        for face in ('x_y', 'x_z', 'y_z'):
            #self.faces[face].set_array(data[face])  # this would be better, but it crashes...
            result.append(self._plot_face(face, data[face]))
            self.data[face] = data[face]  # <- updated for bookkeeping/consistency.
        return result

    # # # ADDING STUFF TO PLOT # # #
    def colorbar(self, **kw_plt_colorbar):
        '''add a colorbar to the figure. sets self.cbar = Colorbar object, and returns it.
        Troubleshooting colors? Try providing vmin, vmax, & levels, explicitly!
        '''
        face = self.faces['x_y']
        kw_plt_colorbar.update(self.plot_settings.get('colorbar_kw', default=dict()))
        self.cbar = self.fig.colorbar(face, **kw_plt_colorbar)
        return self.cbar


@pcAccessor.register('faceplot')
@PlotSettings.format_docstring()
class Faceplot(MoviePlotNode):  # [TODO] refactor PlotDimsMixin to use it here.
    '''MoviePlotNode of a faceplot.
    stores a FaceplotPlotElement, and updates it for each frame of the movie.
    "faceplot" refers to 3  plots on 3 faces of a box in 3D.

    Troubleshooting colors? Try providing vmin, vmax, & levels, explicitly!

    data: dictlike with keys 'x_y', 'x_z', 'y_z', or an xarray.DataArray.
        2D arrays of data to plot on the faces. E.g. 'x_z' data goes on the x-z face.
        xarray.DataArray --> infer faces via isel z=0, y=0, and x=0, respectively.
    coords: None or dictlike with keys 'x', 'y', 'z'
        1D coordinate arrays along each dimension.
        If None, infer from data['x'], data['y'], data['z'].
    t: None, str, or int
        the array dimension which frames will index. E.g. 'time'.
        None --> infer it via infer_movie_dim(data['x_y'].dims, t).
        str --> name of time dimension. And, data should be xarray object(s).
        int --> index of time dimension. And, data should be dict of numpy arrays.

    init_plot: {init_plot}

    add_colorbar: {add_colorbar}
    add_labels: {add_labels}

    title: {title}
    title_font: {title_font}
    title_y: {title_y}
    title_kw: {title_kw}
    '''
    element_cls = FaceplotPlotElement

    def __init__(self, data, *, coords=None, t=None, ax=None,
                 init_plot=PlotSettings.default('init_plot'),  # <- could go in kw_super, but explicit is nice.
                 add_colorbar=PlotSettings.default('add_colorbar'),
                 add_labels=PlotSettings.default('add_labels'),
                 **kw_super):
        if isinstance(data, xr.DataArray):
            data = xarray_faces_3D(data)
        xyarr = data['x_y']
        if hasattr(xyarr, 'dims') and xyarr.ndim > 2:
            t = infer_movie_dim(xyarr.dims, t)
        if (xyarr.ndim != 2) and (t is None):
            errmsg = 't=None but is required (when array.ndim != 2). But cannot infer t from data provided.'
            raise PlottingAmbiguityError(errmsg)
        self.t = t
        if coords is None:
            coords = {x: data[x] for x in 'xyz'}
        self.coords = coords
        self.data = data
        self._ax_init = ax
        kw_super.update(init_plot=init_plot, add_colorbar=add_colorbar, add_labels=add_labels)
        super().__init__(**kw_super)

    # # # PROPERTIES # # #
    ax = alias_child('obj', 'ax', doc='''mpl_toolkits.mplot3d.axes3d.Axes3D where this Faceplot is plotted.''')
    @ax.getter
    def ax(self):  # ax=None if not self.plotted
        return self.obj.ax if self.plotted else None

    fig = alias_child('obj', 'fig', doc='''figure where this Faceplot is plotted.''')
    @fig.getter
    def fig(self):  # fig=None if not self.plotted
        return self.obj.fig if self.plotted else None

    # # # METHODS FOR PLOTTING (parent class wants these to be implemented here) # # #
    def init_plot(self):
        '''plot for the first time. Save the FaceplotPlotElement at self.obj.'''
        self._init_plot_checks()
        frame = self.plot_settings['init_plot_frame']
        data = self.get_data_at_frame(frame)
        # get settings for plot
        kw_plot = self.plot_settings.get_mpl_kwargs('pc.FaceplotPlotElement')
        # -- determine vlims (based on the entire dataset, not just this frame)
        kw_plot = update_vlims([self.data[face] for face in ('x_y', 'x_z', 'y_z')], kw_plot)
        # >> actually make the plot on self.ax <<
        self.obj = self.element_cls(self.coords, data, ax=self._ax_init, **kw_plot)
        # -- if add_labels, add title:
        if self.plot_settings['add_labels']:
            self.plot_title()
        # bookkeeping
        self.frame = frame

    def get_data_at_frame(self, frame):
        '''returns dict-like of data at this frame, ready for plotting.'''
        t = self.t
        result = self.data
        if t is not None:  # there is a dimension for the time axis.
            if isinstance(result, xr.Dataset):
                result = xarray_isel(result, {t: frame})
            else:
                result = dict()
                for face in ('x_y', 'x_z', 'y_z'):
                    if isinstance(result[face], xr.DataArray):
                        result[face] = xarray_isel(result[face], {t: frame})
                    else:
                        result[face] = result[face].take(frame, axis=t)
        return result

    def get_nframes_here(self):
        '''return the number of frames that could be in the movie, based on this node.'''
        t = self.t
        if t is None:
            return 1
        # else, there is a time dimension. return its length:
        data = self.data
        if isinstance(data, xr.Dataset):
            result = len(data[t])
        else:
            if isinstance(data['x_y'], xr.DataArray):
                result = len(data['x_y'][t])
            else:
                result = data['x_y'].shape[t]
        return result

    # # # TITLE # # #
    def plot_title(self):
        '''adds title (as a MovieTextNode) on self.ax.
        raise PlottingAmbiguityError if title already plotted
            (this prevents having multiple title nodes).
        '''
        if hasattr(self, 'title_node'):
            raise PlottingAmbiguityError(f'{type(self).__name__} title was already plotted!')
        title = self.plot_settings['title']
        if title is not None:
            kw_title = {**self.plot_settings.kw, 'init_plot': True}
            title_node = xarray_title_plot_node(self.data['x_y'], title, t=self.t,
                                                ax=self.ax, parent=self, **kw_title)
            self.title_node = title_node  # <-- for bookkeeping/debugging



"""
File Purpose: caching into files / loading cached values from files

[TODO] calculator.behavior.serialize() would help avoid errors.
Implementation here is somewhat primitive.
"""
import os
import shutil

import numpy as np
import xarray as xr

from ..quantity_loader import QuantityLoader
from ...errors import (
    InputError, InputMissingError, MemorySizeError,
    CacheNotApplicableError,
)
from ...tools import (
    simple_property,
    format_docstring,
    code_snapshot_info,
    xarray_save, xarray_load,
)
from ...defaults import DEFAULTS


class CachesLoader(QuantityLoader):
    '''caching into files / loading cached values from files.
    Ideal for relatively small arrays which are expensive to calculate.
        (e.g. ln_std_blur_deltafrac_n, the blurring can take a bit of time for large blur_sigma.)
    For large arrays, recommended: just use xarray_save and xarray_load, directly.

    Stores relevant behavior_attrs and never gives a wrong answer when loading from cache.
        if self.behavior doesn't match stored behavior attrs, delete cached value.
        if current PlasmaCalcs version doesn't match version noted in array.attrs, delete cached value.
        E.g. can use self('caches_ln_std_blur_deltafrac_n') without worrying about blur_sigma;
            if current blur_sigma differs from stored blur_sigma, deletes cache instead of using it.
            And if default blur_sigma changes (via a change to PlasmaCalcs), the version will be different,
                so the cache will also be deleted. (checks __version__ and git commit hash if available.)
    
    Caution: may delete files with '_pc_caches' in their path, without warning.
    '''

    cache_dirname = simple_property('_cache_dirname', setdefaultvia='_default_cache_dirname',
        doc='''abspath to directory for containing cached values of this CachesLoader.
        Default: {self.dirname}/_pc_caches if self.dirname exists, else raise InputMissingError.
        Caution: PlasmaCalcs may delete files with '_pc_caches' in their path, without warning.''')

    def _default_cache_dirname(self):
        '''default for self.cache_dirname.
        {self.dirname}/_pc_caches if self.dirname exists, else raise InputMissingError.
        '''
        dirname = getattr(self, 'dirname', None)
        if isinstance(dirname, str):
            return os.path.join(dirname, '_pc_caches')
        else:
            errmsg = f'self.cache_dirname not set, and got non-str self.dirname={self.dirname}.'
            raise InputMissingError(errmsg)

    def _assert_cache_array_not_too_big(self, array):
        '''checks that array is not too big for caching. If too big, crash with MemorySizeError.'''
        mbmax = DEFAULTS.CACHE_ARRAY_MBYTES_MAX
        if (mbmax is not None) and (array.pc.nMbytes > mbmax):
            errmsg = (f'Array too big for caching ({array.pc.nMbytes} MB); limit={mbmax} MB.\n'
                      f'To adjust limit, set DEFAULTS.CACHE_ARRAY_MBYTES_MAX (currently {mbmax}).')
            raise MemorySizeError(errmsg)

    def _cache_dst(self, array_or_name):
        '''return abspath for cached array, given array (or name of array).
        (if DataArray, uses array.name; if Dataset, must provide name as str instead.)
        result will include .pcxarr extension.

        if inferred name includes '/', crash (and suggest to use ÷ character instead).
        '''
        if isinstance(array_or_name, str):
            name = array_or_name
        else:
            name = getattr(array_or_name, 'name', None)
        if name is None:
            raise InputError(f'cannot infer name; consider providing str directly. Got: {array_or_name}')
        if '/' in name:
            raise InputError("cannot get _cache_dst for name containing '/'. Consider using '÷' instead.")
        result = os.path.join(self.cache_dirname, f'{name}.pcxarr')
        return result

    def _delete_cache_dst_if_exists(self, dst):
        '''delete dst, if it exists, but first ensure it contains '_pc_caches' in the path.
        Never deletes files which don't contain '_pc_caches' in the path.
        '''
        if os.path.exists(dst):
            if '_pc_caches' in dst:
                if os.path.isdir(dst):  # folder
                    shutil.rmtree(dst)
                else:  # file
                    os.remove(dst)
            else:
                raise InputError(f"Refusing to auto-delete file without '_pc_caches' in path: {dst!r}")

    def _array_compatible_with_current_behavior(self, array, *, return_info=False):
        '''returns whether array is compatible with current self.behavior.

        Compatible if all of the following are true:
        (1) All attrs appearing in array.attrs & self.behavior must be equal.
                E.g., if 'units' in both, need array.attrs['units'] == self.behavior['units'].
        (2) All attrs appearing in self.behavior.nondefault(include_xr=False) must appear in array.attrs.
                E.g., if 'blur_sigma' in self.behavior.nondefault(), need 'blur_sigma' in array.attrs.
        (3) [TODO] If self.behavior.nondefault(include_xr='only') has any values,
            they must match values saved to arrays in os.path.join(array.attrs['filepath'], 'behavior_arrays')
                E.g., if 'tfbi_grid1_zeros' appears, need self.behavior['tfbi_grid1_zeros'].equals(behavarr),
                where behavarr = xarray_load(... the array at 'behavior_arrays/tfbi_grid1_zeros.pcxarr'))
            [TODO] currently, caching behavior array values is not implemented;
                see type(self).caches_behavior_skip_xr for details.
        (4) All appearing in array and self.behavior.dims must be equal.
                E.g., if 'fluid' in both, need array_equal(array['fluid'].values, self.behavior.dims['fluid'])
        (5) If array.attrs has 'pc__version__' and/or 'pc__commit_hash',
            they must match the current PlasmaCalcs version and commit hash, if known.

        return_info: bool
            whether to instead return (result, reason of mismatch, key, array value, behavior value).
            (reason will be 'version', 'attrs', 'dims', or 'array_attrs')
            (if result is True, all other values are None)
        '''
        behavior = self.behavior
        # (5)  # <-- check first for efficiency
        if 'pc__version__' in array.attrs or 'pc__commit_hash' in array.attrs:
            code_info = code_snapshot_info()
            for k in ['pc__version__', 'pc__commit_hash']:
                if k in array.attrs and k in code_info and array.attrs[k] != code_info[k]:
                    if return_info:
                        return False, 'version', k, array.attrs[k], code_info[k]
                    else:
                        return False
        # (1)
        for k, v in array.attrs.items():
            if k in behavior:
                if behavior[k] != v:
                    if return_info:
                        return False, 'attrs', k, v, behavior[k]
                    else:
                        return False
        # (2)
        for k in behavior.nondefault_keys(ql=self, include_xr=(not self.caches_behavior_skip_xr)):
            if k not in array.attrs:
                if return_info:
                    return False, 'attrs', k, None, behavior[k]
                else:
                    return False
        # (3)  # [TODO]
        # for k, v in behavior.nondefault(ql=self, include_xr='only').items():
        #     if 'filepath' not in array.attrs:
        #         if return_info:
        #             return False, 'array_attrs', k, None, object.__repr__(v)
        #         else:
        #             return False
        #     filepath = array.attrs['filepath']
        #     behavpath = os.path.join(filepath, 'behavior_arrays', f'{k}.pcxarr')
        #     if not os.path.exists(behavpath):
        #         if return_info:
        #             return False, 'array_attrs', k, None, object.__repr__(v)
        #         else:
        #             return False
        #     behavarr = xarray_load(behavpath)
        #     if not v.equals(behavarr):
        #         if return_info:
        #             return False, 'array_attrs', k, object.__repr__(behavarr), object.__repr__(v)
        #         else:
        #             return False
        # (4)
        for k, v in behavior.dims.items():
            if k in array.coords:
                if not np.array_equal(array[k].values, v):
                    if return_info:
                        return False, 'dims', k, array[k].values, v
                    else:
                        return False
        if return_info:
            return True, None, None, None, None
        else:
            return True

    def _assert_array_compatible_with_current_behavior(self, array):
        '''assert that array is compatible with current self.behavior.
        (if not, crash with CacheNotApplicableError)
        '''
        result, kind, key, arr_val, behav_val = self._array_compatible_with_current_behavior(array, return_info=True)
        if not result:
            errmsg = f'Cache {kind!r} mismatch for {key!r}:\n(>>cache<<) {arr_val}\n(>>current<<) {behav_val}.'
            raise CacheNotApplicableError(errmsg)

    cls_behavior_attrs.register('caches_behavior_skip_xr', default=False)
    caches_behavior_skip_xr = simple_property('_caches_behavior_skip_xr', default=False,
        doc='''where to skip array-valued behavior attrs when caching arrays (to _pc_caches)
        and when checking compatibility with already-cached arrays (in _pc_caches).
        CAUTION: if True, might give subtly incorrect results if the relevant array-valued behavior attrs change.
        Eventually, caching should save array-valued attrs too, but it's trickier so this is a workaround for now.''')

    @known_pattern(r'caches_(.+)', deps=[0])
    @format_docstring(default_mbmax=DEFAULTS.CACHE_ARRAY_MBYTES_MAX)
    def get_caches_var(self, var, *, _match=None):
        '''get var, possibly from folder within self.cache_dirname.
        puts self.behavior.nondefault() as attrs of result.
            (caching not yet implemented for arrays with complicated nondefault behavior,
            e.g. will fail if trying to use caches_var syntax and masking at the same time.
            Workaround: separately call xarray_save and xarray_load, for results you care about.
            Other workaround: use )

        if previously cached, check if applicable, i.e.
            whether attrs agree with self.behavior.nondefault(), and dims agree with self.behavior.dims.
            if applicable, then return cached result.
            else, destroy cache and proceed as if not already cached (see below).

        if not already cached, saves to {{self.cache_dirname}}/{{var}},
            unless result is too big (in which case, crash with MemorySizeError).
            "too big" [MB] == DEFAULTS.CACHE_ARRAY_MBYTES_MAX (default: {default_mbmax})

        see also: cached_var, cache_var
        '''
        here, = _match.groups()
        dst = self._cache_dst(here)
        if os.path.exists(dst):
            # get from cache
            result = xarray_load(dst)
            if self._array_compatible_with_current_behavior(result):
                return result
            else:
                # cache is incompatible; delete it and proceed as if not cached.
                self._delete_cache_dst_if_exists(dst)
        # not cached yet. get, then save to cache
        result = self(here, assign_behavior_attrs=True,
                      assign_behavior_attrs_skip_xr=self.caches_behavior_skip_xr,
                      assign_behavior_attrs_max_call_depth=2+self.call_depth)
        self._assert_cache_array_not_too_big(result)
        for attr, v in result.attrs.items():
            if isinstance(v, (xr.DataArray, xr.Dataset)):
                errmsg = (f'Cannot cache result when behavior attr {attr!r},\n'
                          f'has nondefault value of type DataArray or Dataset (got: {type(v)}).\n'
                          'Consider using self.caches_behavior_skip_xr=True as a workaround for now.')
                raise CacheNotApplicableError(errmsg)
        xarray_save(result, dst)  # might crash if complicated behavior attrs.
        return result

    @known_pattern(r'cached_(.+)', deps=[0])
    def get_cached_var(self, var, *, _match=None):
        '''get var from folder within self.cache_dirname.
        if not already cached, crash with CacheNotApplicableError.
        if previously cached, ensures attrs agree with self.behavior.nondefault(),
            and ensures dims agree with self.behavior.dims.

        see also: caches_var, cache_var
        '''
        here, = _match.groups()
        dst = self._cache_dst(here)
        if not os.path.exists(dst):
            raise CacheNotApplicableError(f'No cache file for {here!r}; checked: {dst!r}')
        result = xarray_load(dst)
        self._assert_array_compatible_with_current_behavior(result)
        return result

    @known_pattern(r'cache_(.+)', deps=[0])
    def get_cache_var(self, var, *, _match=None):
        '''get var, then save to cache. returns value of var.
        if cached result already exists, will overwrite it.
        if result is too big, crash with MemorySizeError.

        see also: caches_var, cached_var
        '''
        here, = _match.groups()
        result = self(here, assign_behavior_attrs=True,
                      assign_behavior_attrs_skip_xr=self.caches_behavior_skip_xr,
                      assign_behavior_attrs_max_call_depth=2+self.call_depth)
        for attr, v in result.attrs.items():
            if isinstance(v, (xr.DataArray, xr.Dataset)):
                errmsg = (f'Cannot cache result when behavior attr {attr!r},\n'
                          f'has nondefault value of type DataArray or Dataset (got: {type(v)}).\n'
                          'Consider using self.caches_behavior_skip_xr=True as a workaround for now.')
                raise CacheNotApplicableError(errmsg)
        self._assert_cache_array_not_too_big(result)
        dst = self._cache_dst(here)
        self._delete_cache_dst_if_exists(dst)
        xarray_save(result, dst)
        return result



"""
File Purpose: Snap, SnapList, SnapDimension, SnapHaver

"snap" is shorthand for "snapshot".
"""
import numpy as np

from .dimension_tools import (
    DimensionValue, DimensionValueList, UniqueDimensionValue,
    DimensionSpecialValueSpecifier,
    Dimension, DimensionHaver,
)
from ..errors import (
    SnapKeyError, SnapValueError, InputError,
    FileContentsConflictError,
    DimensionalityError,
)
from ..tools import (
    elementwise_property,
    xarray_assign,
    format_docstring,
    NO_VALUE, UNSET,
    ArraySelectableChildHaver,
    unique_close, is_integer,
    DictlikeFromKeysAndGetitem,
)
from ..defaults import DEFAULTS


''' --------------------- Snap & SnapList --------------------- '''

class Snap(DimensionValue):
    '''snap... the label (str) and index (int) of a snapshot.

    The "index" should only be meaningful in the context of a SnapList.
    The "label" should be the str name for this snapshot
        - unique within context (e.g. there's only one "snapshot 1" in a simulation)
        - easiest to use str int labels (e.g. "snapshot 1" --> label="1")

    s: the label (str) of the snapshot. if None, cannot convert to str.
    i: the "index" (int) of the snapshot (within a SnapList). if None, cannot convert to int.
    t: time at this snapshot ['raw' units]
    '''
    # attrs "defining" this DimensionValue, & which can be provided as kwargs in __init__
    _kw_def = {'s', 'i', 't'}

    def __init__(self, s=None, i=None, *, t=None):
        super().__init__(s, i)
        self.t = t

    def copy_add_t(self, t_add):
        '''create a copy of self but with t increased by t_add.
        Equivalent to self.copy(t=self.t + t_add).
        Provided separately for convenience; shifting t can be quite useful.
        '''
        return self.copy(t=self.t + t_add)

    def __repr__(self):
        contents = [repr(val) for val in [self.s, self.i] if val is not None]
        if self.t is not None:
            contents.append(f't={self.t:.3e}')
        return f'{type(self).__name__}({", ".join(contents)})'

    @classmethod
    def from_dict(self, d, *, ignore_dt=True):
        '''return cls from dict d. d should NOT have 'typename' key.
        Use cls.deserialize(d) instead if d has 'typname' key.

        ignore_dt: bool
            whether to ignore 'dt' key from d.
            This exists for backwards compatibility when reading older files;
            Snap previously accepted 'dt' key but doesn't anymore.
        '''
        d_use = {**d}
        if ignore_dt:
            d_use.pop('dt', None)
        return super().from_dict(d_use)

    def file_snap(self, calculator):
        '''returns the snap used when determining file related to this snap.
        Here, just returns self. Subclasses might override.
        '''
        return self

    def file_s(self, calculator):
        '''returns the str used when determining file related to this snap.
        Here, just return str(self). subclass should override if desired.
        '''
        return str(self)

    def file_path(self, calculator):
        '''returns the abspath to the file associated with this snap for this calculator.
        Here, just return calculator.snap_filepath(self).
        '''
        return calculator.snap_filepath(self)

    def exists_for(self, calculator):
        '''returns whether this snap exists for this calculator.
        Here, just return self is not MISSING_SNAP. subclass should override if desired.
        '''
        return self is not MISSING_SNAP


class SnapList(DimensionValueList,
               ArraySelectableChildHaver, array_selectable_child='t'):
    '''list of snaps'''
    _dimension_key_error = SnapKeyError
    value_type = Snap

    @classmethod
    def from_lists(cls, s, *, t=None):
        '''return SnapList from zipping these lists. (i will be determined automatically.)'''
        try:
            iter(t)
        except TypeError:
            t = [t] * len(s)
        return cls.from_dicts([dict(s=s_, t=t_) for s_, t_, in zip(s, t)])

    @classmethod
    def from_array(cls, array):
        '''return SnapList from 0D or 1D array.
        array values should be Snap objects, strings, or ints.
            Snap objects --> use as-is
            strings --> use Snap(s, i=i) for i in range(len(array))
            ints --> use Snap(i=i) for i in array
        if array is a DataArray with 't' coord and any value is not a Snap,
            use the corresponding t value when creating the Snap objects.
        '''
        # [TODO] use super() instead of repeating code from DimensionValueList.from_array?
        if array.ndim == 0:
            values = [array.item()]
        elif array.ndim == 1:
            values = array.values
        else:
            errmsg = f'{cls.__name__}.from_array expects array.ndim=0 or 1; got ndim={array.ndim}'
            raise DimensionalityError(errmsg)
        t_size_matches = False
        if 't' in array.coords:
            tt = array.coords['t']
            if tt.size == len(values):
                t_size_matches = True
        result = []
        for i, v in enumerate(values):
            if isinstance(v, cls.value_type):
                result.append(v)
            elif isinstance(v, str):
                kw_t = dict(t=tt[i]) if t_size_matches else dict()
                result.append(cls.value_type(v, i=i, **kw_t))
            elif is_integer(v):
                kw_t = dict(t=tt[i]) if t_size_matches else dict()
                result.append(cls.value_type(i=v, **kw_t))
            else:
                errmsg = (f'{cls.__name__}.from_array got unexpected value type at index {i}.\n'
                            f'Expected {cls.value_type.__name__}, str, or int; got value={v!r}.')
                raise InputError(errmsg)
        return cls(result)

    t = elementwise_property('t', as_array=True)

    def copy_add_t(self, t_add):
        '''create a copy of self but with t values for all snaps increased by t_add'''
        new_snaps = [snap.copy_add_t(t_add) for snap in self]
        return type(self)(new_snaps)

    def file_s(self, calculator):
        '''return array of file_s for each snap in self, for this calculator.
        Here, just returns array of [snap.file_s(calculator) for snap in self].
        '''
        return np.array([snap.file_s(calculator) for snap in self])

    def file_snap(self, calculator):
        '''return array of file_snap for each snap in self, for this calculator.
        Here, just returns array of [snap.file_snap(calculator) for snap in self])'''
        return np.array([snap.file_snap(calculator) for snap in self])

    def file_path(self, calculator):
        '''returns array of file_path for each snap in self, for this calculator.
        Here, just returns array of [snap.file_path(calculator) for snap in self].
        '''
        return np.array([snap.file_path(calculator) for snap in self])

    def exists_for(self, calculator):
        '''return array of snap.exists_for(calculator) for each snap in self.'''
        return np.array([snap.exists_for(calculator) for snap in self])

    def existing_snaps(self, calculator):
        '''return list like self but skipping any snaps where snap.file_snap(calculator) is MISSING_SNAP.
        The length of this result might be different from the length of self.
        '''
        exists = self.exists_for(calculator)
        keepsnaps = [s for s, exist in zip(self, exists) if exist]
        return type(self)(keepsnaps)


class UniqueSnap(UniqueDimensionValue, Snap):
    '''unique snap; Sentinel value with Snap type.'''
    def copy_add_t(self, _t_add):
        '''return self, unchanged; self is a Sentinel so there is only 1 instance.'''
        return self


INPUT_SNAP = UniqueSnap('INPUT_SNAP')
MISSING_SNAP = UniqueSnap('MISSING_SNAP')


class SnapSelectorValueSpecifier(DimensionSpecialValueSpecifier):
    '''class to specify SnapDimension values based on a selector.
    E.g. snaps.get(SELECT_BETWEEN(0.1, 0.5)) <-- equivalent --> snaps.select_betwee(0.1, 0.5).
    E.g. (snap_dim.v = SELECT_AFTER(3.7)) <--> (snap_dim.v = snap_dim.values.select_after(3.7))
    
    getter: str
        self specifies to use the value(s): snaps.{getter}(*args_getter, **kw_getter).
        Internally stored at self.init_getter.
    name: str
        name of this object; will only be used in repr.
    args_getter: None or iterable
        args to pass to snap_dim.values.getter
    kw_getter: None or dict-like
        kwargs to pass to snap_dim.values.getter
    '''
    def __init__(self, getter, name=None, *, args_getter=None, kw_getter=None):
        self.init_getter = getter
        self.name = name
        self.args_getter = args_getter
        self.kw_getter = kw_getter

    def _new(self, **kw_init):
        '''make a new SnapSelectorValueSpecifier like self, with same init_getter and maybe same name.'''
        result = type(self)(self.init_getter, **kw_init)
        if result.name is None: result.name = self.name
        return result

    def __repr__(self):
        if self.name is None:
            return f'{type(self).__name__}({self._getter_as_str})'
        else:
            return self.name

    def __call__(self, *args_getter, **kw_getter):
        '''returns a new SnapSelectorValueSetter with updated args and kwargs.'''
        return self._new(args_getter=args_getter, kw_getter=kw_getter)

    def getter(self, dlist):
        '''returns dlist.{getter}(args_getter, kw_getter).
        if self.args_getter or self.kw_getter not provided, crash with InputError.
        '''
        if self.args_getter is None or self.kw_getter is None:
            typestr = type(self).__name__
            errmsg = f"{typestr} requires args_getter or kw_getter to be provided."
            if self.name is not None:
                errmsg += (f"\nMaybe forgot to call {typestr} instance?"
                           f"\nE.g. instead of using {self.name}, use {self.name}(...)")
            raise InputError(errmsg)
        values_getter = getattr(dlist, self.init_getter)
        return values_getter(*self.args_getter, **self.kw_getter)

SELECT_CLOSEST = SnapSelectorValueSpecifier('select_closest', 'SELECT_CLOSEST')
SELECT_BEFORE = SnapSelectorValueSpecifier('select_before', 'SELECT_BEFORE')
SELECT_AFTER = SnapSelectorValueSpecifier('select_after', 'SELECT_AFTER')
SELECT_ALL_BEFORE = SnapSelectorValueSpecifier('select_all_before', 'SELECT_ALL_BEFORE')
SELECT_ALL_AFTER = SnapSelectorValueSpecifier('select_all_after', 'SELECT_ALL_AFTER')
SELECT_BETWEEN = SnapSelectorValueSpecifier('select_between', 'SELECT_BETWEEN')


''' --------------------- SnapDimension, SnapHaver --------------------- '''

@format_docstring(dimdoc=Dimension.__doc__)
class SnapDimension(Dimension, name='snap', plural='snaps',
                    value_error_type=SnapValueError, key_error_type=SnapKeyError):
    '''snap dimension, representing current value AND list of all possible values.
    Also has various helpful methods for working with this Dimension.

    here, defines custom assign_coord & assign_coord_along,
    because snap is associated with 't' coord.

    u: None or UnitsManager
        used by assign_coord to determine units for t coord.
        if None, will not assign t coord.

    See help(Dimension) for more details.

    --- Dimension.__doc__ copied below: ---
    {dimdoc}
    '''
    NAN = np.nan  # value to use for data when snap is MISSING_SNAP

    def __init__(self, *, u=None, **kw_super):
        super().__init__(**kw_super)
        self.u = u

    def assign_coord(self, array, value=NO_VALUE, **kw_xarray_assign):
        '''returns array.assign_coords(...), assigning 'snap' AND 't' coords.

        value: NO_VALUE, int, str, or Snap
            if not NO_VALUE, assign snap coord as if self.snap = value.
            (set self.snap=value, assign_snap_coord, then restore original self.snap)

        requires not self.snap_is_iterable(), else raises SnapValueError.

        if self.snap.t does not exist,
            raise SnapValueError if DEFAULTS.DEBUG, else do not assign 't' coord.
            if 't' coord is unexpectedly not being assigned, investigate via DEFAULTS.DEBUG = True.

        Note: 't' represents physical time, while 'snap' includes snap name information.

        units for 't' will be self.u('t', alt='coords') units when this method is called.
            assumes snap.t is in 'raw' units; will use:
            t = snap.t * self.u('t', alt='coords')
        '''
        if value is not NO_VALUE:
            with self.using(v=value):
                return self.assign_coord(array, **kw_xarray_assign)
        if self.is_iterable():
            raise self.dimension_value_error("SnapDimension.assign_coord requires non-iterable self.v")
        # else, actually assign the coord.
        snap = self.v
        t_raw = getattr(snap, 't', None)
        u = getattr(self, 'u', None)
        if u is None or t_raw is None:  # don't assign t coord.
            if DEFAULTS.DEBUG and t_raw is None:
                raise self.dimension_value_error(f"SnapDimension.assign_coord requires snap.t to exist. snap={snap}")
            if DEFAULTS.DEBUG and u is None:
                raise self.dimension_value_error(f"SnapDimension.assign_coord requires self.u to exist. self={self}")
            return xarray_assign(array, coords={'snap': snap}, **kw_xarray_assign)
        # else, t and u are both provided. So, assign t coord t, too.
        t = t_raw * u('t', alt='coords')  # time in [u 'coords' units system]
        return xarray_assign(array, coords={'snap': snap, 't': t}, **kw_xarray_assign)

    def assign_coord_along(self, array, dim, value=NO_VALUE, **kw_assign_coords):
        '''returns array.assign_coords(...), assigning 'snap' AND 't' coords.

        value: NO_VALUE, None, or iterable indicating which snaps to use.
            if not NO_VALUE, assign snap coord as if self.snap = value.
            (set self.snap=value, assign_snap_coord, then restore original self.snap)

        requires self.snap_is_iterable(), else raises SnapValueError.

        if self.snap.t does not exist,
            raise SnapValueError if DEFAULTS.DEBUG, else do not assign 't' coord.
            if 't' coord is unexpectedly not being assigned, investigate via DEFAULTS.DEBUG.

        Note: 't' represents physical time, while 'snap' includes snap name information.

        units for 't' will be self.units when this method is called.
            assumes snap.t is in 'raw' units; will use:
            t = snap.t * self.u('t', self.units, convert_from='raw')
        '''
        # [TODO] encapsulate repeated code from previous method.
        if value is not NO_VALUE:
            with self.using(v=value):
                return self.assign_coord_along(array, dim)
        if not self.is_iterable():
            raise self.dimension_value_error("SnapDimension.assign_coord_along requires iterable self.v")
        # else, actually assign the coord.
        snap = self.v
        u = getattr(self, 'u', None)
        t_raw = getattr(snap, 't', None)
        if t_raw is None:  # maybe snap is an iterable of Snap objects
            t_raw = []
            for s in snap:
                st = getattr(s, 't', None)
                if st is None:
                    t_raw = None
                    break
                t_raw.append(st)
            else:  # didn't break
                t_raw = np.array(t_raw)
        if u is None or t_raw is None:  # don't assign t coord.
            if DEFAULTS.DEBUG and t_raw is None:
                raise self.dimension_value_error(f"SnapDimension.assign_coord requires snap.t to exist. snap={snap}")
            if DEFAULTS.DEBUG and u is None:
                raise self.dimension_value_error(f"SnapDimension.assign_coord requires self.u to exist. self={self}")
            return array.assign_coords({'snap': (dim, snap)}, **kw_assign_coords)
        # else, t and u are both provided. So, assign t coord t, too.
        t = t_raw * u('t')  # time in [self.u units system]
        return array.assign_coords({'snap': (dim, snap), 't': (dim, t)}, **kw_assign_coords)

    def join_along(self, arrays, *, coords=['t'], **kw_super):
        '''return arrays joined along the snap dimension.
        by default, joins 't' coords as well.

        coords: str or list of str
            names of coords to join along with snap.
            default = ['t'] --> join 't' coords as well.
            str --> rule, e.g. 'different' or 'minimal'; see help(xr.concat) for details.
        '''
        if not isinstance(coords, str):
            if len(coords)==1 and coords[0]=='t':
                if not all('t' in array.coords for array in arrays):
                    coords = UNSET
        cdict = dict() if coords is UNSET else dict(coords=coords)
        return super().join_along(arrays, **cdict, **kw_super)

    def n_existing_for(self, calculator):
        '''return number of snaps which exist for this calculator.
        Equivalent to (self.values.exists_for(calculator)).sum()
        '''
        return self.values.exists_for(calculator).sum()

    def current_n_existing_for(self, calculator):
        '''return number of existing snaps, out of all snaps at self.v.
        Equivalent to self.v.exists_for(calculator).
        '''
        # note, self.v might not be a SnapList.
        result = 0
        for snap in self.iter():  # iter through v, not values.
            if snap.exists_for(calculator):
                result += 1
        return result


@SnapDimension.setup_haver
class SnapHaver(DimensionHaver, dimension='snap', dim_plural='snaps'):
    '''class which "has" a SnapDimension. (SnapDimension instance will be at self.snap_dim)
    self.snap stores the current snap (possibly multiple). If None, use self.snaps instead.
    self.snaps stores "all possible snaps" for the SnapHaver.
    Additionally, has various helpful methods for working with the SnapDimension,
        e.g. current_n_snap, iter_snaps, take_snap.
        See SnapDimension.setup_haver for details.

    Also, indexing self will set self.snap, then return self.
        see help(self.__getitem__) for more details.
    '''
    def __init__(self, *, snap=None, snaps=None, **kw):
        super().__init__(**kw)
        if snaps is not None: self.snaps = snaps
        self.snap = snap
        self.snap_dim.u = self.u

    def __getitem__(self, snapi):
        '''sets self.snap, then returns self.
        Examples:
            self[4] --> self.snap = 4
            self[:5] --> self.snap = slice(0, 5)
            self[-1] --> self.snap = -1
            self['2'] --> self.snap = '2'
            self[self.snaps[7]] --> self.snap = self.snaps[7]
            self[[0,4,7,3]] --> self.snap = [0,4,7,3]
        Note this is 'smart' in the same way that setting self.snap is 'smart';
            will attempt self.snaps.get(snapi).
            See help(self.snaps.get) for more details.

        This enables "shorthand" syntax, e.g. self[3]('n') gets 'n' at self.snap=3.
        '''
        self.snap = snapi
        return self

    def __iter__(self):
        '''equivalent to self.iter_snaps()'''
        return self.iter_snaps()

    def __len__(self):
        '''equivalent to len(self.snaps)'''
        return len(self.snaps)

    def n_existing_snaps(self):
        '''returns number of existing snaps. Equivalent to self.snap_dim.n_existing_for(self).'''
        return self.snap_dim.n_existing_for(self)

    def current_n_existing_snaps(self):
        '''returns number of existing snaps, out of all snaps at self.snap.
        Equivalent to self.snap_dim.current_n_existing_for(self).
        '''
        return self.snap_dim.current_n_existing_for(self)

    def existing_snaps(self):
        '''return list of existing snaps. Equivalent to self.snaps.existing_snaps(self).'''
        return self.snaps.existing_snaps(self)

    # # # FILE MANAGEMENT -- SUBCLASS SHOULD IMPLEMENT # # #
    @property
    def snapdir(self):
        '''directory containing the snapshot files. Subclass should implement.'''
        raise NotImplementedError(f'{type(self).__name__}.snapdir property')

    def snap_filepath(self, snap=None):
        '''convert snap to full file path for this snap. Subclass should implement.

        snap: None, str, in, or Snap
            the snapshot to load. if None, use self.snap.
        '''
        raise NotImplementedError(f'{type(self).__name__}.snap_filepath()')


''' --------------------- ProxySnap --------------------- '''

_proxy_id_registry = dict()  # {proxy_id: calculator} pairs

class ProxySnap(Snap):
    '''snap which is a proxy for one point in time but possibly different snap indices for different runs.
    proxies: dict or None
        {proxy_id(calculator): snap} pairs indicating the snap for each calculator.
        if None, make an empty dict.
    additional args & kwargs are passed to super().__init__.
    '''
    _dimension_value_error = SnapValueError

    def __init__(self, s=UNSET, i=None, *, proxies=None, **kw_super):
        self.proxies = dict() if proxies is None else proxies
        super().__init__(s=s, i=i, **kw_super)

    @property
    def s(self):
        '''s associated with this ProxySnap. if UNSET, infer from self.proxies.values().
        if all proxies values have the same s, use it. Else, use None.
        '''
        if self._s is UNSET:
            return self._s_from_proxies()
        else:
            return self._s
    @s.setter
    def s(self, value):
        self._s = value
    @s.deleter
    def s(self):
        self._s = UNSET

    def _s_from_proxies(self):
        '''return common s from proxies, if any. Else, return None.'''
        s = None
        for snap in self.proxies.values():
            if s is None:
                s = snap.s
            elif s != snap.s:
                s = None
                break
        return s

    def file_snap(self, calculator):
        '''returns the snap which self.proxies points at for this calculator.
        equivalent to self.proxies[self.proxy_id(calculator)].
        '''
        return self.proxies[self.proxy_id(calculator)]

    def file_s(self, calculator):
        '''returns the str used when determining file related to this snap.
        equivalent to self.proxies[self.proxy_id(calculator)].s
        '''
        snap = self.file_snap(calculator)
        return snap.s

    def exists_for(self, calculator):
        '''returns whether this snap exists for this calculator.
        equivalent to self.proxies[self.proxy_id(calculator)].exists_for(calculator).
        '''
        snap = self.file_snap(calculator)
        return snap.exists_for(calculator)

    @classmethod
    def proxy_id(cls, calculator):
        '''return the unique id used to identify this calculator in a ProxySnap.
        unique id guaranteed to be hashable.
        '''
        try:
            return calculator._proxy_id  # this calculator already has a unique id.
        except AttributeError:
            pass  # handled below
        cls._assign_proxy_id(calculator)  # assign proxy id then return it.
        return calculator._proxy_id

    @classmethod
    def _assign_proxy_id(cls, calculator):
        '''assign unique calculator._proxy_id.'''
        assert not hasattr(calculator, '_proxy_id'), f"calculator already has a _proxy_id: {calculator._proxy_id}"
        pid = cls._get_next_proxy_id()
        calculator._proxy_id = pid
        _proxy_id_registry[pid] = calculator  # id registry is saved outside of cls to avoid circular dependencies.
    
    @classmethod
    def _get_next_proxy_id(cls):
        '''return the next unique proxy_id.'''
        try:
            result = cls._next_proxy_id
        except AttributeError:
            result = 0
        cls._next_proxy_id = result + 1
        return result

    @staticmethod
    def registry():
        '''returns dict of all defined {proxy_id: calculator} pairs.
        (To avoid circular dependencies, this dict is stored outside of the ProxySnap class,
            but can be looked up here when requested.)
        '''
        return _proxy_id_registry

    @classmethod
    def from_calculators(cls, calculators, t, *, mode='isclose', missing_ok=False, **kw_init):
        '''return ProxySnap with proxies for each calculator at time t.
        t: number
            time, same units as snaps from calculators.
        mode: 'isclose' or 'exact'
            how to determine which snap to use, if any.
            'isclose' --> np.isclose(snap.t, t). 'exact' --> snap.t == t.
        missing_ok: bool
            if any snap matches are not found, tells whether to raise error or use MISSING_SNAP.
            True --> raise SnapValueError. False --> use MISSING_SNAP for any missing snaps.
        '''
        proxies = {}
        force_relative = {'isclose': 'isclose', 'exact': 'equal'}[mode]
        for calculator in calculators:
            snaps = calculator.snaps
            try:
                snap = snaps.select_closest(t, force_relative=force_relative)
            except ValueError:
                if missing_ok:
                    snap = MISSING_SNAP
                else:
                    raise
            # avoid using multiple layers of proxy, if unnecessary.
            if isinstance(snap, ProxySnap) and cls.proxy_id(calculator) in snap.proxies:
                snap = snap.proxies[cls.proxy_id(calculator)]
            # set proxy for this calculator:
            proxies[cls.proxy_id(calculator)] = snap
        return cls(proxies=proxies, t=t, **kw_init)


class ProxySnapList(SnapList):
    '''list of ProxySnaps'''
    value_type = ProxySnap

    @classmethod
    def from_calculators(cls, calculators, *, mode='isclose', join='inner', **kw_init):
        '''return ProxySnapList of all snaps relevant when considering all calculators.
        In case of different snap lists, behavior is determined by kwargs.

        mode: 'isclose' or 'exact'
            how to determine snap matching.
            'isclose' --> np.isclose(snap.t, t). 'exact' --> snap.t == t.
        join: 'inner' or 'outer'. [TODO] implement other join options?
            how to determine which snaps to keep.
            'inner' --> keep only snaps with matching t.
            'outer' --> keep all snaps (use MISSING_SNAP as needed to fill gaps).
        '''
        if len(calculators)==0:
            return cls()
        if join == 'inner':
            # start with the shortest list of snaps; attempt to get ProxySnap for each t.
            snap_lists = [calculator.snaps for calculator in calculators]
            snaps = min(snap_lists, key=len)
            times = snaps.t
            result = []
            i = 0
            for t in times:
                try:
                    proxy = ProxySnap.from_calculators(calculators, t, mode=mode, i=i)
                except ValueError:  # snap match is missing from one of the calculators.
                    continue
                else:
                    result.append(proxy)
                    i += 1
        elif join == 'outer':
            # get list of all times from all calculators. Use MISSING_SNAP where necessary.
            times = np.unique(np.concatenate([calculator.snaps.t for calculator in calculators]))
            # account for "close" values up to floating point errors (except for at 0. Use atol=0.)
            times = unique_close(times, atol=0)
            times = np.sort(times)
            result = []
            for i, t in enumerate(times):
                proxy = ProxySnap.from_calculators(calculators, t, mode=mode, i=i, missing_ok=True)
                result.append(proxy)
        else:
            raise ValueError(f"join={join!r} not supported.")
        return cls(result, **kw_init)

    def set_in_calculators(self, calculators, *, reset_current_snap=True):
        '''sets calculator.snaps = self for each calculator in calculators.
        Note this sets snaps, not snap. So, the list of "all possible values" is changed.

        reset_current_snap: bool
            whether to set c.snap = 0 or None, for c in calculators.
            if False, c.snap will remain unchanged.
            Default to using None, but if c.snap == 0, use 0 instead.
        '''
        for c in calculators:
            c.snaps = self
            if reset_current_snap:
                if c.snap == 0:
                    c.snap = 0  # set it to the NEW 0 (from self), not the old 0 (from old snaps)
                else:
                    c.snap = None  # equivalent to c.snap = c.snaps


''' --------------------- ParamsSnap --------------------- '''

class ParamsSnap(DictlikeFromKeysAndGetitem, Snap):
    '''single snapshot with associated parameters, saved in self.params;
    self also behaves like a dict with params keys & values.
    (Useful if each snapshot has a possibly-different set of parameters.)

    s: the label (str) of the snapshot. if None, cannot convert to str.
    i: the "index" (int) of the snapshot (within a SnapList). if None, cannot convert to int.
    t: time at this snapshot ['raw' units]
    params: dict of parameters at this snapshot. if None, make empty dict.

    The "index" should only be meaningful in the context of a SnapList
    The "label" should be the str name for this snapshot
        - unique within context (e.g. there's only one "snapshot 1" in a simulation)
        - easiest to use str int labels (e.g. "snapshot 1" --> label="1")
    '''
    def __init__(self, s=None, i=None, *, t=None, params=None):
        params = dict() if params is None else params
        super().__init__(s=s, i=i, t=t)
        self.params = params

    # # # GETTING VALUES / ITERATING # # #
    def __getitem__(self, key):
        '''equivalent to self.params[key]'''
        return self.params[key]

    def keys(self):
        '''return tuple of keys of self.params'''
        return tuple(self.params.keys())


class ParamsSnapList(SnapList):
    '''list of ParamsSnap objects. Also provides helpful methods:

    keys(): list of keys appearing in any snap
    params(): dict of {key: [list of values of key from all snaps])
    params_global(): dict of {key: value} for all keys which have the same value in all snaps.
    params_varied(): dict of {key: [list of values of key from all snaps]),
                        for all keys in keys() but not in params_global().
    iter_param_values(key): iterating across all values of key from all snaps,
                        but only yield value once if key is in params_global().
    '''
    value_type = ParamsSnap

    # # # GETTING KEYS & PARAMS # # #
    def keys(self, *, missing_ok=True, recalc=False):
        '''return list of all keys from snaps. Maintains order in which keys appear.

        missing_ok: bool or None
            whether it is okay for snaps to have different keys.
            False --> not okay. If any snap has any different keys, raise FileContentsConflictError.
            True --> okay. If any snap has different keys, add to list of keys.
        recalc: bool
            whether to, if possible, return cached value (not a copy -- CAUTION: don't edit directly!)
        '''
        # caching
        if not recalc:
            if hasattr(self, '_keys_same'):  # result of self.keys(missing_ok=False)
                # if this result exists, then missing_ok doesn't matter either way;
                # because it didn't crash we know all the snaps have the same keys.
                return self._keys_same
            elif missing_ok and hasattr(self, '_keys_join'):  # result of self.keys(missing_ok=True)
                return self._keys_join
        # computing result
        result = []
        if len(self) == 0:
            return result
        result = list(self[0].keys())
        result_set = set(result)
        for snap in self[1:]:
            if missing_ok:
                for key in snap.keys():
                    if key not in result:
                        result.append(key)
            else:
                # ensure key set here matches key set in result, else crash.
                snap_key_set = set(snap.keys())
                if snap_key_set != result_set:
                    missing_here = result_set - snap_key_set
                    missing_zero = snap_key_set - result_set
                    errmsg = (f'Some snaps have different keys. '
                              f'Keys in this snap but not snap 0: {missing_here}. '
                              f'Keys in snap 0 but not this snap: {missing_zero}. '
                              f'This snap = {snap}')
                    raise FileContentsConflictError(errmsg)
        # caching
        if missing_ok:
            self._keys_join = result
        else:
            self._keys_same = result
        return result

    def params(self, *, missing_ok=True, recalc=False):
        '''return dict of {key: [list of values of key from all snaps])
        
        missing_ok: bool
            whether it is okay for snaps to have different keys.
            False --> not okay. If any snap has any different keys, raise FileContentsConflictError.
            True --> okay. When key is missing from a snap, use NO_VALUE instead.
        recalc: bool
            whether to, if possible, return cached value (not a copy -- CAUTION: don't edit directly!)
        '''
        # caching
        if not recalc:
            if hasattr(self, '_params_same'):  # result of self.params(missing_ok=False)
                # if this result exists, then missing_ok doesn't matter either way;
                # because it didn't crash we know all the snaps have the same keys.
                return self._params_same
            elif hasattr(self, '_params_join') and missing_ok:  # result of self.params(missing_ok=True)
                return self._params_join
        # computing result
        keys = self.keys(missing_ok=missing_ok, recalc=recalc)
        result = dict()
        if missing_ok:
            for key in keys:
                result[key] = [snap.get(key, NO_VALUE) for snap in self]
        else:
            for key in keys:
                result[key] = [snap[key] for snap in self]
        # caching
        if missing_ok:
            self._params_join = result
        else:
            self._params_same = result
        return result

    def keys_shared(self, *, recalc=False):
        '''return list of keys which appear in all snaps. Maintains order in which keys appear.

        recalc: bool
            whether to, if possible, return cached value (not a copy -- CAUTION: don't edit directly!)
        '''
        # caching
        if not recalc and hasattr(self, '_keys_shared'):
            return self._keys_shared
        # computing result
        result = []
        if len(self) == 0:
            return result
        result = list(self[0].keys())
        result_set = set(result)
        for snap in self[1:]:
            snap_key_set = set(snap.keys())
            extras_here = result_set - snap_key_set
            for extra in extras_here:
                result.remove(extra)
                result_set.remove(extra)
        # caching
        self._keys_shared = result
        return result

    def params_global(self, *, recalc=False):
        '''return dict of {key: value} for all keys which have the same value in all snaps.

        recalc: bool
            whether to, if possible, return cached value (not a copy -- CAUTION: don't edit directly!)
        '''
        # caching
        if not recalc and hasattr(self, '_params_global'):
            return self._params_global
        # computing result
        if len(self) == 0:
            return dict()
        elif len(self) == 1:
            return self[0].params.copy()
        # else len(self) > 1
        keys = self.keys_shared(recalc=recalc)
        keys = list(keys)  # make a copy to avoid directly editing possibly cached result.
        result = {key: self[0][key] for key in keys}
        for snap in self[1:]:
            for key in keys:
                if snap[key] != result[key]:
                    del result[key]
                    keys.remove(key)
        # caching
        self._params_global = result
        return result

    def keys_global(self, *, recalc=False):
        '''return list of keys whose value is the same in all snaps. Maintains order in which keys appear.

        recalc: bool
            whether to, if possible, return cached value (not a copy -- CAUTION: don't edit directly!)
        '''
        # caching
        if not recalc and hasattr(self, '_keys_global'):
            return self._keys_global
        # computing result
        keys = self.keys_shared(recalc=recalc)
        keys = list(keys)  # make a copy to avoid directly editing possibly cached result.
        params_global = self.params_global(recalc=recalc)
        keys_shared_set = set(keys)
        keys_global_set = set(params_global.keys())
        keys_remove_set = keys_shared_set - keys_global_set
        for key in keys_remove_set:
            keys.remove(key)
        # caching
        self._keys_global = keys
        return keys

    def keys_varied(self, *, recalc=False):
        '''return list of keys whose value is different in any snap. Maintains order in which keys appear.

        recalc: bool
            whether to, if possible, return cached value (not a copy -- CAUTION: don't edit directly!)
        '''
        # caching
        if not recalc and hasattr(self, '_keys_varied'):
            return self._keys_varied
        # computing result
        keys_all = self.keys(missing_ok=True, recalc=recalc)
        keys_global = self.keys_global(recalc=recalc)
        keys_varied = [key for key in keys_all if key not in keys_global]
        # caching
        self._keys_varied = keys_varied
        return keys_varied

    def params_varied(self, *, recalc=False):
        '''return dict of {key: [values from all snaps]) for all params with different values in any snap.
        If any snaps are missing any keys, use NO_VALUE as the value.

        recalc: bool
            whether to, if possible, return cached value (not a copy -- CAUTION: don't edit directly!)
        '''
        # caching
        if not recalc and hasattr(self, '_params_varied'):
            return self._params_varied
        # computing result
        keys = self.keys_varied(recalc=recalc)
        result = dict()
        for key in keys:  # looping like this makes debugging easier than list comprehension.
            result[key] = [snap.get(key, NO_VALUE) for snap in self]
        # caching
        self._params_varied = result
        return result

    def iter_param_values(self, key, default=UNSET):
        '''iterate over all values of this param from all snaps.
        If value is the same in all snaps, just yield it once then stop.
        If value is missing from all snaps, yield default if provided else crash.

        default: UNSET or any object
            for snaps missing key, yield default if provided else crash.
        '''
        params_global = self.params_global()
        if key in params_global:
            yield params_global[key]
        else:  # key not global
            keys = self.keys()
            if key in keys:  # key exists in at least 1 snap
                for snap in self:
                    if key in snap.keys():
                        yield snap[key]
                    elif default is UNSET:
                        raise FileContentsConflictError(f'key {key!r} missing from snap {snap}')
                    else:
                        yield default
            else:  # key not in any snaps
                if default is UNSET:
                    raise FileContentsConflictError(f'key {key!r} missing from all snaps.')
                else:
                    yield default

    def list_param_values(self, key, default=UNSET):
        '''list all values of this param from all snaps.
        If value is the same in all snaps, return [value]
        If value is missing from all snaps, return [default] if provided else crash.

        default: UNSET or any object
            for snaps missing key, use default if provided else crash.
        '''
        return list(self.iter_param_values(key, default=default))


"""
File Purpose: calculating stats especially relevant to plasma
"""

from .patterns import StatsLoader

class PlasmaStatsLoader(StatsLoader):
    '''plasma stats, e.g. mean weighted by n.'''
    @known_pattern(r'nmean_(.+)', deps=[0])
    def get_nmean(self, var, *, _match=None):
        '''mean of var, weighted by n. nmean = sum(var * n) / sum(n). n=self('n').
        Equivalent: 'nmean_var' <--> 'weighted_n_mean_var'
        '''
        var, = _match.groups()
        return self(f'weighted_n_mean_{var}')

    @known_pattern(r'nstd_(.+)', deps=[0])
    def get_nstd(self, var, *, _match=None):
        '''std of var, weighted by n. nstd = std(n*var)/mean(n), with n=self('n').'''
        var, = _match.groups()
        return self(f'weighted_n_std_{var}')



"""
Package Purpose: PlasmaCalculator for Eppic
"""
from .eppic_bases import EppicBasesLoader
from .eppic_calculator import EppicCalculator
from .eppic_choose_params import (
    EppicParam, EppicChooseParamsBase,
    EppicChooseParams,
    EppicTimescales,
)
from .eppic_dimensions import (
    EppicDist, EppicDistList,
    EppicNeutral, EppicNeutralList,
)
from .eppic_direct_loader import EppicDirectLoader
from .eppic_input_deck import EppicInputDeck
from .eppic_instability_calculator import (
    EppicInstabilityCalculator,
    EppicTfbiCalculator, EppicCalculatorWithTfbi,
)
from .eppic_io_tools import (
    attempt_literal_eval, read_eppic_i_file,
    update_eppic_i_file, get_updated_eppic_i_str,
    read_eppic_snaps_info,
    read_moments_out_files,
    infer_eppic_dist_names,
    n_mpi_processors,
    eppic_clock_times_from_jobfile, eppic_clock_times_here,
    read_timers_dat,
)
from .eppic_moments import EppicMomentsLoader
from .eppic_plotters import EppicPlotterManager
from .eppic_runtime_info import EppicRuntimeInfoLoader
from .eppic_sim_info import EppicSimInfoLoader
from .eppic_subsampler import EppicSubsamplable
from .eppic_hybrid_calculator import EppicHybridCalculator


"""
File Purpose: misc. tools related to QuantityLoader
"""

import functools
import re

import numpy as np

from ..errors import InputError, InputConflictError, InputMissingError
from ..tools import (
    IncrementableAttrManager,
    alias, alias_child,
    format_docstring,
    UNSET,
    PartitionFromXarray, xarray_promote_dim, xarray_squeeze,
)
from ..defaults import DEFAULTS


_paramdocs = {
    'fname': '''str
        name of the callable used to calculate this quantity,
        via getattr(obj,fname)(...), where obj is a QuantityLoader.''',
    'var': '''str
        the var matched to this object.''',
    '_match': '''result of re.match
        the result of pattern.fullmatch(var) for the pattern associated with this object.
        None --> no pattern match; either no associated pattern or var is not a match.''',
    'deps': '''list of str/int/tuple/dict/callable
        list of dependencies for this quant. Each dependency is one of the following:
            - a str, indicating a known var or pattern name,
            - an int, indicating a group index from a pattern match,
            - a 2-tuple of (non-tuple-dep, dict),
                where dict tells any values essential to set via quantity_loader.using(**dict),
                    in order to properly determine deps.
                E.g. ('n', {'fluid': ELECTRON}) for 'depends on n when self.fluid=electron'.
                Note: mostly intended for internal use (in value_deps). It's usually preferable to
                    just define a different variable, e.g. 'ne' which will set values appropriately.
            - a dict with length 1, with
                key = int or tuple of ints, specifying which groups are relevant.
                        dep will utilize info about groups = _match.groups().
                val = str, or callable f -> str or iterable.
                        str --> val.format(**group_info), where
                                group_info = {f'group{i}': groups[i] for i in key},
                                    replacing groups[i] with '' if groups[i] is None.
                        callable f --> f(group_info), where
                                group_info = {i: groups[i] for i in key}
            - a callable of the form f(quantity_loader, var, groups) -> str or iterable,
                which tells the var name(s) associated with this dependency,
                where: quantity_loader is a QuantityLoader instance,
                    var is the matched var,
                    and groups=_match.groups() if _match provided else None.''',
    'attr_deps': '''list of 2-tuples of (str, dict or str)
        list of attr-based dependencies for this quant. Each dependency is:
            (attr, lookup), where lookup.get(v, []) tells str or list of deps associated with v,
            where v=quantity_loader.attr.
            if '__default__' in lookup, use lookup.get(v, lookup['__default__']) instead.
            str lookup --> lookup = getattr(quantity_loader, lookup).''',
    'value_deps': '''list of 2-tuples of (str, dict or str)
        list of value-based dependencies for this quant. Each dependency is:
            (var, lookup), where lookup.get(val, []) tells str or list of deps associated with val,
            for all unique vals in quantity_loader(var).
            Each dep associated with val may be str/int/callable, similar to `deps`.
            if '__default__' in lookup, use lookup.get(v, lookup['__default__']) instead.
            str lookup --> lookup = getattr(quantity_loader, lookup).''',
    'dims': '''None or list of strings
        dimensions associated directly with this quantity. E.g. ['fluid', 'snap']
        None --> this quant is not directly associated with any dims.''',
    'ignores_dims': '''list of strings
        dimensions which the result ignores, even if the dependencies don't ignore.
        E.g. 'mod_(.+)' has ignores_dims=['component'],
            because it provides the same result regardless of obj.component.''',
    'reduces_dims': '''list of strings
        dimensions which the result reduces along.
        E.g. 'ldebye_subset' has reduces_dims=['fluid'],
            because the result doesn't have 'fluid' dim even though it depends on obj.fluid.'''
    }


''' --------------------- LoadableQuantity/Var/Pattern --------------------- '''

@format_docstring(**_paramdocs)
class LoadableQuantity():
    '''quantity which can be loaded from a QuantityLoader.
    QuantityLoader.KNOWN_VARS and KNOWN_PATTERNS will be a dict of {{str: LoadableQuantity}} pairs.

    fname: {fname}
    deps: {deps}
    attr_deps: {attr_deps}
    value_deps: {value_deps}
    dims: {dims}
    ignores_dims: {ignores_dims}
    reduces_dims: {reduces_dims}
    '''
    def __init__(self, fname, *, deps=[], attr_deps=[], value_deps=[],
                 dims=None, ignores_dims=[], reduces_dims=[], **kw_super):
        self.fname = fname
        self.deps = deps
        self.attr_deps = attr_deps
        self.value_deps = value_deps
        if len(attr_deps)>0 and isinstance(attr_deps[0], str):
            raise InputError(f'attr_deps should be list of 2-tuples of (str, dict), but got {attr_deps!r}.')
        if len(value_deps)>0 and isinstance(value_deps[0], str):
            raise InputError(f'value_deps should be list of 2-tuples of (str, dict), but got {value_deps!r}.')
        self.dims = dims
        self.ignores_dims = ignores_dims
        self.reduces_dims = reduces_dims
        super().__init__(**kw_super)

    # # # DICT-LIKE # # #
    def keys(self):
        '''returns keys whose values can be accessed via self[key].
        i.e.: ('fname', 'deps', 'attr_deps', 'value_deps', 'dims', 'ignores_dims', 'reduces_dims').
        '''
        return ('fname', 'deps', 'attr_deps', 'value_deps', 'dims', 'ignores_dims', 'reduces_dims')

    def items(self):
        '''returns (key, self[key]) for key in self.keys().'''
        return ((key, getattr(self, key)) for key in self.keys())

    def __getitem__(self, key):
        '''returns self.{key}, if key in self.keys().'''
        if key not in self.keys():
            raise KeyError(f'{key!r} not in self.keys()={self.keys()}.')
        return getattr(self, key)

    # # # GETTING THE CALLABLE TO USE # # #
    def get_f(self, quantity_loader):
        '''returns the callable to use to calculate this quantity, for this quantity_loader.'''
        # note: methods outside of this class should not assume other methods will use get_f;
        #   they should feel free to use getattr(quantity_loader, loadable_quantity.fname).
        # However, inside this class, methods should use get_f instead of getattr(...).
        return getattr(quantity_loader, self.fname)

    def get_f_module(self, quantity_loader):
        '''returns the module in which this quantity is defined, for this quantity_loader.'''
        return self.get_f(quantity_loader).__module__

    # # # DISPLAY # # #
    def _repr_deps(self):
        '''return repr for deps in self.'''
        result = []
        for dep in self.deps:
            if callable(dep):
                result.append(f'callable: {dep.__name__}')
            else:
                result.append(repr(dep))
        return '[' + ', '.join(result) + ']'

    def _repr_special_deps(self, special_deps):
        '''return repr for attr_deps or value_deps in self.'''
        result = []
        for dep in special_deps:
            special, lookup = dep
            if isinstance(lookup, str):
                lookup_copy = lookup
            else:
                lookup_copy = {k: v for k, v in lookup.items()}
                for k, v in lookup_copy.items():
                    if callable(v):
                        lookup_copy[k] = f'callable: {v.__name__}'
                    else:
                        lookup_copy[k] = v
            result.append(f'({special!r}, {lookup_copy!r})')
        return '[' + ', '.join(result) + ']'  # e.g. [('attr', {'val1': 'dep', 'val2': [0, 'dep2']})]

    def _repr_contents(self):
        '''return contents for __repr__.'''
        contents = [f'{self.fname!r}']
        if len(self.deps) > 0:
            contents.append(f'deps={self._repr_deps()}')
        if len(self.attr_deps) > 0:
            contents.append(f'attr_deps={self._repr_special_deps(self.attr_deps)}')
        if len(self.value_deps) > 0:
            contents.append(f'value_deps={self._repr_special_deps(self.value_deps)}')
        if self.dims is not None:
            contents.append(f'dims={self.dims!r}')
        if len(self.ignores_dims) > 0:
            contents.append(f'ignores_dims={self.ignores_dims!r}')
        if len(self.reduces_dims) > 0:
            contents.append(f'reduces_dims={self.reduces_dims!r}')
        return contents

    def __repr__(self):
        return f'{type(self).__name__}({", ".join(self._repr_contents())})'


@format_docstring(**_paramdocs)
class LoadableVar(LoadableQuantity):
    '''quantity which can be loaded from a QuantityLoader.
    QuantityLoader.KNOWN_VARS will be a dict of {{str: LoadableVar}} pairs.

    fname: {fname}
    deps: {deps}
        (Note: for LoadableVar, deps will not include ints, since there won't be any pattern match.)
    attr_deps: {attr_deps}
    value_deps: {value_deps}
    dims: {dims}
    ignores_dims: {ignores_dims}
    reduces_dims: {reduces_dims}
    '''
    pass  # all functionality inherited from LoadableQuantity.


@format_docstring(**_paramdocs)
class LoadablePattern(LoadableQuantity):
    '''quantity which can be loaded from a QuantityLoader.
    QuantityLoader.KNOWN_PATTERNS will be a dict of {{Pattern: LoadablePattern}} pairs.

    fname: {fname}
    deps: {deps}
    attr_deps: {attr_deps}
    value_deps: {value_deps}
    dims: {dims}
    ignores_dims: {ignores_dims}
    reduces_dims: {reduces_dims}
    '''
    pass  # all functionality inherited from LoadableQuantity.


''' --------------------- LoadQuantity/Var/Pattern --------------------- '''

@format_docstring(**_paramdocs)
class MatchedQuantity():
    '''LoadableQuantity matched to a var, which can be loaded from a QuantityLoader.

    var: {var}
    loadable: LoadableQuantity or None
        the LoadableQuantity associated with this MatchedQuantity instance.
        None --> create one from fname, deps, dims, ignores_dims, and reduces_dims,
                using defaults if UNSET. (however, fname must be provided in this case.)
    _match: {_match}

    self.fname aliases to self.loadable.fname.
    similarly for deps, attr_deps, value_deps, dims, ignores_dims, reduces_dims.
    also similarly for self.get_f and self.get_f_module.
    '''
    def __init__(self, var, loadable=None, *, _match=None,
                 fname=None, deps=UNSET, attr_deps=UNSET, value_deps=UNSET,
                 dims=UNSET, ignores_dims=UNSET, reduces_dims=UNSET,
                 **kw_super):
        self.var = var
        if loadable is None:
            if fname is None:
                raise InputMissingError("'fname' is required when loadable=None.")
            kw_loadable = dict(deps=deps, attr_deps=attr_deps, value_deps=value_deps,
                               dims=dims, ignores_dims=ignores_dims, reduces_dims=reduces_dims)
            kw_loadable = {k: v for k, v in kw_loadable.items() if v is not UNSET}
            loadable = LoadableQuantity(**kw_loadable)
        self.loadable = loadable
        self._match = _match
        super().__init__(**kw_super)

    fname = alias_child('loadable', 'fname')
    deps = alias_child('loadable', 'deps')
    attr_deps = alias_child('loadable', 'attr_deps')
    value_deps = alias_child('loadable', 'value_deps')
    dims = alias_child('loadable', 'dims')
    ignores_dims = alias_child('loadable', 'ignores_dims')
    reduces_dims = alias_child('loadable', 'reduces_dims')
    get_f = alias_child('loadable', 'get_f')
    get_f_module = alias_child('loadable', 'get_f_module')

    def groups(self, *, default=UNSET):
        '''return self._match.groups().
        default: UNSET or any object
            if _match is None, return default if provided (not UNSET), else raise TypeError.
        '''
        if self._match is None:
            if default is UNSET:
                raise TypeError(f'self.groups() when self._match=None, and default=UNSET.')
            else:
                return default
        return self._match.groups()

    def _list_attr_deps(self, quantity_loader=None, *, missing_ok=False):
        '''returns list of deps, based on quantity_loader and self.attr_deps.
        Each dep in result is a str, int, tuple, dict, or callable.
        quantity_loader must be provided if any attr_deps are provided.
        missing_ok: bool
            whether to return [] for attr_dep when quantity_loader is a class and attr is a property.
            False --> crash with InputError in that case.
        '''
        result = []
        if len(self.attr_deps) > 0 and quantity_loader is None:
            raise InputMissingError(f'quantity_loader, when attr_deps are provided. Got attr_deps={self.attr_deps}.')
        for attr, lookup in self.attr_deps:
            val = getattr(quantity_loader, attr)
            if isinstance(lookup, str):
                lookup = getattr(quantity_loader, lookup)
            if (not missing_ok) and isinstance(val, property) and isinstance(quantity_loader, type):
                errmsg = (f'from attr_deps... attr={attr!r} is a property,\n'
                          f'and quantity_loader is a class ({quantity_loader.__name__}),\n'
                          f'Cannot appropriately get deps in this case (when var={self.var!r}).\n'
                          f'Consider using an instance of {quantity_loader.__name__} instead of the class itself,\n'
                          f'or consider setting missing_ok=True to ignore this attr_dep.')
                raise InputError(errmsg)
            default = lookup.get('__default__', [])
            deps = lookup.get(val, default)
            if isinstance(deps, str):
                deps = [deps]
            result.extend(deps)
        return result

    def _list_value_deps(self, quantity_loader=None):
        '''returns list of deps, based on quantity_loader and self.value_deps.
        Each dep in result is a str, int, tuple, dict, or callable.
        quantity_loader must be provided if any value_deps are provided.

        Note: includes the "values" as deps too, not just the lookup dicts.
        E.g. if value_deps = [('var0': {'val0': 0, 'val1': [1, 'b']})],
            and quantity_loader('var0') gives xr.DataArray(['val1', 'val1']),
            result here will be ['var0', 1, 'b'].
        '''
        result = []
        if len(self.value_deps) > 0:
            if quantity_loader is None:
                raise InputMissingError(f'quantity_loader, when value_deps are provided. Got value_deps={self.value_deps}.')
            elif isinstance(quantity_loader, type):
                errmsg = (f'quantity_loader is a class ({quantity_loader.__name__}),\n'
                          f'and value_deps are provided. Cannot appropriately get deps in this case.\n'
                          f'Consider using an instance of {quantity_loader.__name__} instead of the class itself.')
                raise InputError(errmsg)
        for var, lookup in self.value_deps:
            result.append(var)
            if isinstance(lookup, str):
                lookup = getattr(quantity_loader, lookup)
            default = lookup.get('__default__', [])
            vals = quantity_loader(var)
            # maybe this is a bit hacky... it's a way to ensure we properly do quantity_loader.using(**using) for deps.
            # [TODO] do it in a way that is slightly less hacky / allows things to be more explicitly specified...?
            if vals.ndim > 1:
                vals = vals.squeeze()
            if any(dimtype in vals.dims for dimtype in getattr(quantity_loader, '_dim_types', ())):
                # assume self.{dim} needs to be set to the values in vals, to properly determine deps.
                # e.g. if vals is an xarray of strs with 'fluid' dim,  assume lookup[val] corresponds
                #      to quantity_loader.using(fluid=(fluids where vals==val))
                if vals.ndim != 1:
                    errmsg = ('[TODO] value_deps with ndim > 1, when at least 1 dim in quantity_loader._dim_types.\n'
                              f'(got var={var!r}, vals.dims={vals.dims!r}, dimtypes={list(quantity_loader._dim_types)})')
                    raise NotImplementedError(errmsg)
                dim = vals.dims[0]
                partition = PartitionFromXarray(vals, dim)
                for key, dimvals in partition:
                    using = {dim: dimvals}  # must use quantity_loader.using(**using) when getting these deps.
                    deps = lookup.get(key, default)
                    if isinstance(deps, str):
                        deps = [deps]
                    deps = [(dep, using) for dep in deps]
                    result.extend(deps)
            else:  # haven't implemented quantity_loader.using(**using) for non-dimtype-related value deps...
                for val in np.unique(vals):
                    deps = lookup.get(val, default)
                    if isinstance(deps, str):
                        deps = [deps]
                    result.extend(deps)
        return result

    def all_deps(self, quantity_loader=None, *, missing_ok=False):
        '''returns list of deps, based on quantity_loader and self.deps, self.attr_deps, and self.value_deps.
        Each dep in result is a str, int, tuple, dict, or callable.
        See self.dep_vars() to convert list of deps to list of strs.
        If any attr_deps or value_deps are provided, quantity_loader must be provided.

        missing_ok: bool
            whether to return [] for attr_dep when quantity_loader is a class and attr is a property.
            False --> crash with InputError in that case.
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        result = []
        result.extend(self.deps)
        result.extend(self._list_attr_deps(quantity_loader=quantity_loader, missing_ok=missing_ok))
        result.extend(self._list_value_deps(quantity_loader=quantity_loader))
        return result

    @format_docstring(**_paramdocs, sub_indent=DEFAULTS.TAB)
    def dep_vars(self, quantity_loader=None, *, missing_ok=False):
        '''returns list of dep var strings or 2-tuples of (str, using)
        These are the var names associated with self.deps, self.attr_deps, and self.value_deps.
        Each string will be a quant that QuantityLoader might load;
            might be a KNOWN_VAR or might be a KNOWN_PATTERN.
            (However, this method doesn't know whether the deps are actually loadable).
        Each 2-tuple has a string like above,
            and a dict of values essential to set via quantity_loader.using(**dict),
            when trying to determine the deps of that string,
            e.g. during quantity_loader.match_var_tree().
            (If it's not essential to set values, use string instead of 2-tuple.)

        self.deps is a {deps}
        self.attr_deps is a {attr_deps}
        self.value_deps is a {value_deps}

        quantity_loader: None or QuantityLoader
            if any dep is callable, quantity_loader must be provided,
            and is used to get string for dep via dep(quantity_loader, self.var, self.groups(default=None))

            if any attr_deps or value_deps are provided, quantity_loaded must be provided,
            and is used to convert to list of deps based on attr or values.
        missing_ok: bool
            whether to be lenient sometimes when missing details that would allow to fully determine deps.
            E.g., if quantity_loader is a class (e.g. dep_vars called inside QuantityLoader.cls_var_tree()),
                and an attr_dep's attr's value is a property, ignore that attr_dep if missing_ok, else crash. 
        '''
        result = []
        alldeps = self.all_deps(quantity_loader=quantity_loader, missing_ok=missing_ok)
        for dep in alldeps:
            if isinstance(dep, tuple):
                if len(dep) != 2:
                    raise InputError(f'tuple dep must have length 2, but got {dep!r}.')
                dep, using = dep
                dep_vars = self._dep_to_strs(dep, quantity_loader)
                for dep_var in dep_vars:
                    result.append((dep_var, using))
            else:
                dep_vars = self._dep_to_strs(dep, quantity_loader)
                result.extend(dep_vars)
        return result

    def _dep_to_strs(self, dep, quantity_loader=None):
        '''convert dep to list of strings (possibly only 1 string),
        following the logic described in self.dep_vars.__doc__.
        '''
        if isinstance(dep, str):
            dep_var = dep
            result = [dep_var]
        elif isinstance(dep, int):
            groups = self.groups()
            dep_var = groups[dep]
            result = [dep_var]
        elif isinstance(dep, dict):
            if len(dep) != 1:
                raise InputError(f'dict dep must have length 1, but got {dep!r}.')
            groups = self.groups()
            key, f = list(dep.items())[0]
            if isinstance(key, int):
                key = (key,)
            group_info = {k: groups[k] for k in key}
            if isinstance(f, str):
                group_info = {f'group{k}': (v if v is not None else '') for k, v in group_info.items()}
                dep_var = f.format(**group_info)
                dep_vars = [dep_var]
            elif callable(f):
                dep_var = f(group_info)
                dep_vars = [dep_var] if isinstance(dep_var, str) else dep_var
            else:
                raise InputError(f'dict dep value must be str or callable, but got {f!r}.')
            result = dep_vars
        elif callable(dep):
            if quantity_loader is None:
                raise InputMissingError(f'quantity_loader, when one or more deps are callable. (Got dep={dep}).')
            groups = self.groups(default=None)
            dep_var = dep(quantity_loader, self.var, groups)
            if isinstance(dep_var, str):
                result = [dep_var]
            else:
                result = dep_var  # iterable dep_var.
        else:
            errmsg = f'type(dep)={type(dep)}. Expected str or int. (when var={self.var!r})'
            raise TypeError(errmsg)
        return result

    # # # LOADING THE VAR # # #
    def load_value(self, quantity_loader, *args, **kw):
        '''actually get the value of this matched var, from the quantity loader.
        uses getter = self.get_f(quantity_loader);
        if _match is None, calls getter(*args, **kw);
        else, calls getter(self.var, *args, _match=_match, **kw).
        '''
        __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
        getter = self.get_f(quantity_loader)
        _match = self._match
        if _match is None:  # MatchedVar, probably from KNOWN_VARS
            result = getter(*args, **kw)
        else:               # MatchedPattern, probably from KNOWN_PATTERNS
            result = getter(self.var, *args, _match=_match, **kw)
        return result

    # # # DICT-LIKE # # #
    def keys(self):
        '''returns keys whose values can be accessed via self[key].
        i.e.: ('var', 'loadable', '_match', 'fname',
                'deps', 'attr_deps', 'value_deps',
                'dims', 'ignores_dims', 'reduces_dims').
        '''
        return ('var', 'loadable', '_match', 'fname',
                'deps', 'attr_deps', 'value_deps',
                'dims', 'ignores_dims', 'reduces_dims')

    def items(self):
        '''returns (key, self[key]) for key in self.keys().'''
        return ((key, getattr(self, key)) for key in self.keys())

    def __getitem__(self, key):
        '''returns self.{key}, if key in self.keys().'''
        if key not in self.keys():
            raise KeyError(f'{key!r} not in self.keys()={self.keys()}.')
        return getattr(self, key)

    # # # DISPLAY # # #
    def _repr_contents(self):
        '''return contents for __repr__'''
        contents = [f'{self.var!r}']
        contents.extend(self.loadable._repr_contents())
        if self._match is not None:
            contents.append(f'groups={list(self.groups())!r}')
        return contents

    def __repr__(self):
        return f'{type(self).__name__}({", ".join(self._repr_contents())})'


@format_docstring(**_paramdocs)
class MatchedVar(MatchedQuantity):
    '''LoadableVar matched to a var, which can be loaded from a QuantityLoader.

    var: {var}
    loadable: LoadableVar or None
        the LoadableVar associated with this MatchedVar instance.
        None --> create one from fname, deps, dims, ignores_dims, and reduces_dims,
                using defaults if UNSET. (however, fname must be provided in this case.)
    _match: {_match}

    self.fname aliases to self.loadable.fname.
    similarly for deps, attr_deps, value_deps, dims, ignores_dims, reduces_dims.
    also similarly for self.get_f and self.get_f_module.
    '''
    pass  # all functionality inherited from MatchedQuantity


@format_docstring(**_paramdocs)
class MatchedPattern(MatchedQuantity):
    '''LoadableVar matched to a var, which can be loaded from a QuantityLoader.

    var: {var}
    loadable: LoadablePattern or None
        the LoadablePattern associated with this MatchedPattern instance.
        None --> create one from fname, deps, dims, ignores_dims, and reduces_dims,
                using defaults if UNSET. (however, fname must be provided in this case.)
    _match: {_match}

    self.fname aliases to self.loadable.fname.
    similarly for deps, attr_deps, value_deps, dims, ignores_dims, reduces_dims.
    also similarly for self.get_f and self.get_f_module.
    '''
    pass  # all functionality inherited from MatchedQuantity


''' --------------------- Pattern --------------------- '''

class Pattern():
    '''stores re.Pattern and represents it nicely. Use self.pattern to get re.Pattern object.
    Use self.str to get str object.
    Intended to be used in an immutable way; don't change pattern or str after creating.

    pattern: str or re.Pattern object
        will be compiled to re.Pattern object if str.
    '''
    # note: ideally, would've just subclassed re.Pattern.
    #   However, that seems to fail due to re.Pattern implementation.
    def __init__(self, pattern):
        if isinstance(pattern, str):
            pattern = re.compile(pattern)
        self.pattern = pattern
        self.str = pattern.pattern

    def match(self, s, *args_match, **kw_match):
        '''returns self.pattern.match(s, *args, **kw). Equivalent to re.match(self.pattern, s, ...).'''
        return self.pattern.match(s, *args_match, **kw_match)

    def fullmatch(self, s, *args_fullmatch, **kw_fullmatch):
        '''returns self.pattern.fullmatch(s, *args, **kw). Equivalent to re.fullmatch(self.pattern, s, ...).'''
        return self.pattern.fullmatch(s, *args_fullmatch, **kw_fullmatch)

    # # # EQUALITY & HASHING # # #
    def __eq__(self, other):
        '''return whether self and other are equal.
        equal if both are Pattern objects with the same str.
        [TODO] also allow pattern == s if isinstance(s,str) and s == pattern.str?
        '''
        if not isinstance(other, Pattern):
            return False
        return self.str == other.str

    def __hash__(self):
        '''return hash for self: hash((type(self), self.str))'''
        return hash((type(self), self.str))
        
    # # # DISPLAY # # #
    def __repr__(self):
        return f'{type(self).__name__}({self.str!r})'


''' --------------------- DecoratingCalcs/Vars/Patterns/Setters --------------------- '''

class DecoratingCalcs():
    '''Decorating functions to put loadable quantities into a dict (self.known_calcs).

    known_calcs: None or dict
        will store known calcs; dict of {name: LoadableQuantity} pairs.

    Not intended for direct use; see DecoratingVars or DecoratingPatterns instead.
    '''
    loadable_quantity_cls = LoadableQuantity  # class used for making new LoadableQuantity instances.

    def __init__(self, known_calcs=None):
        if known_calcs is None: known_calcs = dict()
        self.known_calcs = known_calcs

    def track_f(self, f, *, name=None, aliases=[],
                deps=[], attr_deps=[], value_deps=[],
                dims=None, ignores_dims=[], reduces_dims=[]):
        '''add f to self.known_calcs as a LoadableQuantity.
        See help(self.decorator) for details on parameters.
        '''
        fname = f.__name__
        if name is None:
            START = 'get_'
            assert fname.startswith(START), f'f.__name__ must start with "{START}"; got {fname!r}.'
            varname = fname[len(START):]
        else:
            varname = name
        lq = self.loadable_quantity_cls(fname, deps=deps, attr_deps=attr_deps, value_deps=value_deps,
                                        dims=dims, ignores_dims=ignores_dims, reduces_dims=reduces_dims)
        if getattr(self, 'cls_associated_with', None) is not None:
            lq.cls_where_defined = self.cls_associated_with
        self.known_calcs[varname] = lq
        for alias_ in aliases:
            self.known_calcs[alias_] = lq

    @format_docstring(**_paramdocs, sub_ntab=1)
    def decorator(self, *, name=None, aliases=[],
                  deps=[], attr_deps=[], value_deps=[],
                  dims=None, load_across_dims=None, ignores_dims=[], reduces_dims=[],
                  partition_across_dim=None, partition_deps=None):
        '''returns decorator for quant calculator f(self, *args, **kw), which returns f unchanged...
        but also sets self.known_calcs[name] = LoadableQuantity(f.__name__, deps=deps)
            (note - actually uses self.loadable_quantity_cls instead of LoadableQuantity)

        name: None or str
            if provided, tells name of this quantity (i.e., the key in self.known_calcs)
            None --> use {{name}} from f.__name__ which looks like "get_{{name}}".
        aliases: list
            if any aliases are provided, also add, for alias in aliases:
                self.known_calcs[alias] = self.known_calcs[name]
        deps: {deps}
        attr_deps: {attr_deps}
        value_deps: {value_deps}
        dims: {dims}
        load_across_dims: None or list of strings
            if provided, the returned decorator(f) will actually wrap f to use load_across_dims:
                result(f)(self, ...) returns self.load_across_dims(f, ..., dims=load_across_dims).
            To indicate dim dependencies without adjusting f, use kwarg `dims`, instead.
        ignores_dims: {ignores_dims}
        reduces_dims: {reduces_dims}
        partition_across_dim: None or 2-tuple of (dim, partitioner)
            if provided, the returned decorator(f) will actually wrap f:
                result(f)(self, ...) partitions across dim and provides partitioner as kwarg input,
                then rejoins along dim afterwards.
                currently, only supports partitioning across 1 dim per f.
            partitioner should be a string, and will be added to deps.
        partition_deps: None, str, or dict
            if provided, tells lookup from partitioner value to additional dep var name(s).
            (only allowed if partition_across_dim is provided.)
            Equivalent to using value_deps=[(partitioner, partition_deps)].
            str --> will use lookup = getattr(quantity_loader, partition_deps).

            E.g. {{'saha': 'ionfrac_saha', 'SINGLE_FLUID': ['ne', 'SF_n']}},
                when partition_across_dim=('fluid', 'ionfrac_type'),
                --> add 'ionfrac_saha' to deps if 'saha' in self('ionfrac_type'),
                    and add 'ne' and 'SF_n' to deps if 'SINGLE_FLUID' in self('ionfrac_type').
            and, this example is equivalent to using value_deps=
                [('ionfrac_type', {{'saha': 'ionfrac_saha', 'SINGLE_FLUID': ['ne', 'SF_n']}})]
        '''
        # check for input conflicts:
        if load_across_dims is not None:
            if dims is not None:
                raise InputConflictError('cannot provide both dims and load_across_dims.')
            if any(d in ignores_dims for d in load_across_dims):
                errmsg = (f'ignores_dims cannot include any of the same dims as load_across_dims, but got: '
                          f'ignores_dims={ignores_dims!r} and load_across_dims={load_across_dims!r}.')
                raise InputConflictError(errmsg)
        if partition_across_dim is not None:
            if dims is not None:
                raise InputConflictError('cannot provide both dims and partition_across_dim.')
            if partition_across_dim[0] in ignores_dims:
                errmsg = (f'ignores_dims including the partitioned dim, is currently not implemented. Got: '
                          f'ignores_dims={ignores_dims!r} and partition_across_dim={partition_across_dim!r}.')
                raise InputConflictError(errmsg)
            if load_across_dims is not None:
                raise InputConflictError('cannot provide both load_across_dims and partition_across_dim.')
        # result (if load_across_dims is None and partition_across_dim is None)
        def return_f_after_some_bookkeeping(f):
            '''returns f, unchanged, after calling self.track_f(f, ...) to add it to self.known_calcs.
            (Also adds aliases to self.known_calcs if provided; see help(DecoratingCalcs) for details)
            '''
            self.track_f(f, name=name,aliases=aliases,
                         deps=deps, attr_deps=attr_deps, value_deps=value_deps,
                         dims=dims, ignores_dims=ignores_dims, reduces_dims=reduces_dims)
            return f
        # return result (immediately, if load_across_dims is None and partition_across_dim is None)
        if load_across_dims is None and partition_across_dim is None:
            return return_f_after_some_bookkeeping
        elif load_across_dims is not None:  # make & return wrapper for f
            dims = load_across_dims
            def direct_decorator(f):
                '''returns a decorator for f, which wraps f to use load_across_dims,
                after calling self.track_f(f, ...)
                '''
                f = return_f_after_some_bookkeeping(f)
                @functools.wraps(f)
                def f_but_load_across_dims(self_, *args, **kw):
                    '''f but using load_across_dims.'''
                    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
                    # self_ gets passed as an arg because f is the method before it is bound to self_.
                    with self_.using(ncpu=1, pool=None):
                        # ncpu=1, pool=None is necessary because f (going to load_across_dims) is not pickleable!
                        # that's because the f going to load_across_dims is the f before applying the decorator,
                        #   but the module-level (or class-level) f is the output of the decorator
                        #   (which is a different object from the f inside here)
                        #   and pickle only works on module-level or class-level objects.
                        return self_.load_across_dims(f, self_, *args, dims=dims, **kw)
                return f_but_load_across_dims
            return direct_decorator
        elif partition_across_dim is not None:  # make & return wrapper for f
            dim, partitioner = partition_across_dim
            if dims is None:
                dims = [dim]
            elif dim not in dims:
                dims = dims + [dim]
            if partition_deps is None and partitioner not in deps:
                deps = deps + [partitioner]
            elif partition_deps is not None:
                value_deps = value_deps + [(partitioner, partition_deps)]
            def direct_decorator(f):
                '''returns a decorator for f, which wraps f to partition across dim,
                after calling self.track_f(f, ...)
                '''
                f = return_f_after_some_bookkeeping(f)
                @functools.wraps(f)
                def f_but_partition_across_dim(self_, *args, **kw):
                    '''f but partitioning across dim. {partitioner} will be provided as kwarg.
                    E.g. if partitioner='ntype', will call f(self_, *args, ntype=value, **kw),
                        with value varying across all unique values of self('ntype').

                    if {partitioner} is provided explicitly, skip all partitioning logic,
                        and instead just use the value provided. This helps with debugging,
                        and efficiency improvement if subclass f uses super().f(...)
                    E.g. if partitioner='ntype', but ntype='elem' was provided as kwarg,
                        then just return f(self_, *args, **kw) immediately.
                    '''
                    __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
                    # self_ gets passed as an arg because f is the method before it is bound to self_.
                    if partitioner in kw:   # assume partitioning logic is being handled elsewhere already!
                        return f(self_, *args, **kw)
                    with self_.using(ncpu=1, pool=None):
                        # ncpu=1, pool=None is necessary because f (internally here) is not pickleable!
                        # that's because the f here is the f before applying the decorator,
                        #   but the module-level (or class-level) f is the output for the decorator
                        #   so the f used internally here is not saved at the class-level and cannot be pickled.
                        # [TODO] maybe partition loading logic below should be encapsulated elsewhere...
                        dimobj = self_.dimensions[dim]  # e.g. self_.fluid_dim
                        squeeze_later = (not dimobj.is_iterable())
                        xarray = self_(partitioner)
                        xarray = xarray_squeeze(xarray, keep=dim, drop=False)
                        partition = PartitionFromXarray(xarray, dim, dimobj.list())
                        result = []
                        for partkey, _dvals in dimobj.iter_partition(partition):
                            # >> actually call f <<
                            rr = f(self_, *args, **{partitioner: partkey}, **kw)
                            if dim not in rr.coords:
                                rr = dimobj.assign_coord(rr, expand_if_iterable=True)
                            rr = xarray_promote_dim(rr, dim)
                            result.append(rr)
                        result = dimobj.join_along(result)
                        if squeeze_later:  # e.g. self_.fluid_dim is not iterable. --> squeeze!
                            result = xarray_squeeze(result, dim, drop=False)
                        else:  # isel back to original order to match self.fluid order.
                            result = result.isel({dim: partition.ridx_flat})
                        return result
                return f_but_partition_across_dim
            return direct_decorator
        else:
            assert False, 'coding error if reached this line'

    def __repr__(self):
        return f'{type(self).__name__}({self.known_calcs})'


class DecoratingVars(DecoratingCalcs):
    '''Decorating functions to put loadable vars into a dict (self.known_calcs).
    (e.g., self.known_calcs will be QuantityLoader.KNOWN_VARS; see MetaQuantTracking for details.)

    known_calcs: None or dict
        will store known calcs; dict of {name: LoadableVar} pairs.

    Example: known_var = DecoratingCalcs(storage_dict)
        @known_var
        def get_var1(...):
            ...
        @known_var(name='coolvar', deps=['var1'])
        def get_var2(...):
            ...
        # at this point, we have:
        #   known_var.known_calcs == {'var1': LoadableQuantity('get_var1'),
        #                             'coolvar': LoadableQuantity('get_var2', deps=['var1'])}.
    '''
    loadable_quantity_cls = LoadableVar  # class used for making new LoadableQuantity instances.
    
    def __call__(self, f=None, *, name=None, deps=[], aliases=[],
                 dims=None, load_across_dims=None, ignores_dims=[], reduces_dims=[],
                 partition_across_dim=None, partition_deps=None, **kw_decorator):
        '''if f is provided, return self.decorator(**kw)(f). Otherwise, return self.decorator(**kw).
        This enables instances to be used as decorators directly, i.e. "@self",
            or used as decorators after providing kwargs, e.g. "@self(name='varname', deps=['var2'])"
        See help(type(self)) for examples.
        '''
        kw = dict(name=name, deps=deps, aliases=aliases, dims=dims,
                  load_across_dims=load_across_dims, ignores_dims=ignores_dims, reduces_dims=reduces_dims,
                  partition_across_dim=partition_across_dim, partition_deps=partition_deps,
                  **kw_decorator)
        if f is None:
            return self.decorator(**kw)
        else:
            return self.decorator(**kw)(f)


class DecoratingPatterns(DecoratingCalcs):
    '''DecoratingVars but for patterns; name should be an re.Pattern (or str; and will apply re.compile()),
    and decorated functions will always be provided a value for kwarg _match=re.match(name, var).

    deps values might contain ints; that indicates the dependency is at that group index.
    E.g. '(.*)_([xyz])' would put deps=[0] since it matches {var}_{x} and depends on {var}.
    '''
    loadable_quantity_cls = LoadablePattern  # class used for making new LoadableQuantity instances.

    def decorator(self, name=None, deps=[], *, aliases=[],
                  dims=None, load_across_dims=None, ignores_dims=[], reduces_dims=[],
                  partition_across_dim=None, partition_deps=None):
        '''decorates f(self, var, *, _match=None) like DecoratingCalcs.decorator, but also:
            - name is treated as a Pattern (re.compile() it if necessary)
            - returned function will determine _match via Pattern.fullmatch, if necessary.
                (This way, the function can be called directly, though that is usually discouraged.)
        '''
        pattern = Pattern(name)
        kw_decorator = dict(deps=deps, aliases=aliases,
                            dims=dims, load_across_dims=load_across_dims,
                            ignores_dims=ignores_dims, reduces_dims=reduces_dims,
                            partition_across_dim=partition_across_dim, partition_deps=partition_deps)
        def direct_decorator(f):
            @super(DecoratingPatterns, self).decorator(name=pattern, **kw_decorator)
            @functools.wraps(f)
            def f_with_match_provided(self_, var, *, _match=None, **kw):
                __tracebackhide__ = DEFAULTS.TRACEBACKHIDE
                if _match is None:
                    _match = pattern.fullmatch(var)
                    if _match is None:
                        # (var doesn't match pattern. Probably called f directly, instead of self(var))
                        raise InputError(f'var={var!r} does not match pattern {pattern!r}.')
                return f(self_, var, _match=_match, **kw)
            return f_with_match_provided
        return direct_decorator

    __call__ = alias('decorator')


class DecoratingSetters():
    '''Decorating functions to put them into a dict (self.known_setters).

    known_setters: None or dict
        will store known settables; dict of {name: callable}

    Example: known_setter = DecoratingCalcs(storage_dict)
        @known_setter
        def set_var1(...):
            ...
        @known_setter(name='coolvar')
        def set_var2(...):
            ...
        @known_setter(aliases=['myalias'])
        def set_var3(...):
            ...
        # at this point, we have known_setter.known_setters ==
        # {'var1': set_var1, 'coolvar': set_var2, 'var3': set_var3, 'myalias': set_var3}.
    '''
    def __init__(self, known_setters=None):
        if known_setters is None: known_setters = dict()
        self.known_setters = known_setters

    def decorator(self, name=None, *, aliases=[]):
        '''returns decorator for quant setter f(self, *args, **kw), which returns f unchanged...
            but also, tracks things appropriately:
                add name to self.known_setters,
        if name is not provided, use {name} from f.__name__ which looks like "set_{name}".

         aliases: list
            if any aliases are provided, also add, for alias in aliases:
                self.known_setters[alias] = self.known_setters[name]
        '''
        def return_f(f):
            '''returns f, unchanged, after adding name to self.setters'''
            if name is None:
                START = 'set_'
                fname = f.__name__
                assert fname.startswith(START), f'f.__name__ must start with "{START}"'
                varname = fname[len(START):]
            else:
                varname = name
            self.known_setters[varname] = f
            for alias in aliases:
                self.known_setters[alias] = f
            return f
        return return_f

    def __call__(self, f=None, *, name=None, **kw_decorator):
        '''if f is provided, return self.decorator(**kw)(f). Otherwise, return self.decorator(**kw).
        This enables instances to be used as decorators directly, i.e. "@self",
            or used as decorators after providing kwargs, e.g. "@self(aliases=['varalias'])"
        See help(type(self)) for examples.
        '''
        if f is None:
            return self.decorator(name, **kw_decorator)
        else:
            return self.decorator(name, **kw_decorator)(f)

    def __repr__(self):
        return f'{type(self).__name__}({self.known_setters})'


''' --------------------- CallDepthMixin --------------------- '''

class CallDepthMixin():
    '''Mixin class for tracking call depth.
    call depth == number of times self has been called inside other calls to self.
    see help(type(self).call_depth) for details.

    subclasses should include "with self._increment_call_depth():" inside of __call__, e.g.:
        def __call__(self, *args, **kw):
            with self._increment_call_depth():
                # do stuff; possibly including calling self again.
    '''
    @property
    def call_depth_manager(self):
        '''stores the value of call_depth, and helps to manage attrs dependent on call_depth value.'''
        try:
            result = self._call_depth_manager
        except AttributeError:
            result = IncrementableAttrManager(self, default=0, step=1)
            self._call_depth_manager = result
        return result

    @property
    def call_depth(self):
        '''depth of the current call to self. depth = number of calls to self from within self.

        E.g., call_depth while calculating gyrofrequency:
            # call_depth == 0, for any code run here (outside any call to self).
            self('gyrof')
                # call_depth == 1, for any code run here (inside 'gyrof' call but not inside deeper calls).
                q = self('q') 
                    # call_depth == 2, for code inside 'q' call.
                mod_B = self('mod_B')
                    # call_depth == 2, for code inside 'mod_B' call.
                    self('B')
                        # call_depth == 3, for code inside 'B' call.
                m = self('m')
                    # call_depth == 2, for code inside 'm' call.
                result = q * mod_B / m

        Cannot be set directly; can only be manipulated via self.call_depth_manager.
        '''
        return int(self.call_depth_manager)

    def _increment_call_depth(self):
        '''context manager for incrementing call_depth.

        use "with self._increment_call_depth():" inside of __call__, e.g.:
            def __call__(self, *args, **kw):
                with self._increment_call_depth():
                    # do stuff; possibly including calling self again.

        Equivalent to self.call_depth_manager.increment()
        '''
        return self.call_depth_manager.increment()        

    def using_at_call_depth(self, depth, **attrs_and_values):
        '''context manager for setting attrs_and_values but only while call_depth == depth.

        E.g.:
            with self.using_at_call_depth(3, verbose=3):
                self('sgyrof')
                # while self.call_depth == 3 inside of this 'with' block, uses self.verbose=3.
                #   but everywhere else, uses original value of verbose.
            # assuming originally verbose=False (or unset), this example will print:
            | | (call_depth=2) get var='q'
            | | (call_depth=2) get var='mod_B'
            | | (call_depth=2) get var='m'
            
            # compare this to simply using self.verbose=3, which would print:
            | (call_depth=1) get var='sgyrof'
            | | (call_depth=2) get var='q'
            | | (call_depth=2) get var='mod_B'
            | | | (call_depth=3) get var='B_dot_B'
            | | | | (call_depth=4) get var='B_xyz'
            | | | | | (call_depth=5) get var='B'
            | | (call_depth=2) get var='m'

        Equivalent to self.call_depth_manager.using_obj_attrs_at(depth, **attrs_and_values)
        '''
        return self.call_depth_manager.using_obj_attrs_at(depth, **attrs_and_values)

    def using_at_next_call_depth(self, **attrs_and_values):
        '''context manager for setting attrs_and_values but only while call_depth == self.call_depth + 1

        Equivalent to self.using_at_call_depth(self.call_depth + 1, **attrs_and_values).
        (Also equivalent to self.call_depth_manager.using_obj_attrs_at_next(**attrs_and_values).)
        '''
        return self.using_at_call_depth(self.call_depth + 1, **attrs_and_values)



"""
File Purpose: misc tools related to tfbi
"""

import xarray as xr

from ..instability_tools import itAccessor
from ...errors import InputError
from ...tools import (
    alias,
    is_integer,
    xarray_vars_lookup_with_defaults,
)
from ...defaults import DEFAULTS


''' --------------------- Helper methods --------------------- '''

def T_from_ldebye(ldebye, *, n, eps0, q, kB):
    '''returns T given ldebye. ldebye**2 = eps0 * kB * T / (n * q**2)
    --> T = ldebye**2 * n * q**2 / (eps0 * kB)

    if ldebye is an xarray.DataArray with a name,
        result name will be 'T_from_{ldebye.name}'.
    '''
    result = ldebye**2 * n * (q**2 / (eps0 * kB))
    if isinstance(result, xr.DataArray) and (ldebye.name is not None):
        result.name = f'T_from_{ldebye.name}'
    return result

@itAccessor.register('T_from_ldebye', totype='dataset')
def xarray_T_from_ldebye(ds, ldebye=None, *, n=None, eps0=None, q=None, kB=None):
    '''return T given ldebye. ldebye**2 = eps0 * kB * T / (n * q**2)
    --> T = ldebye**2 * n * q**2 / (eps0 * kB)

    ds: xarray.Dataset
        dataset to use for inputs. Must contain ldebye data.
    ldebye, n, eps0, q, kB: None, str, or value
        None --> use default lookups (see below) for variable names from ds.
        str --> name of variable in ds representing this physical quantity
        value --> directly use this value.

    quantity meanings:
        ldebye: debye length
        n: number density (of each fluid, if multiple fluids)
        eps0: vacuum permittivity
        q: charge (of each fluid, if multiple fluids with different charges)
        kB: boltzmann constant
    default lookups (if not provided explicitly, use first available):
        ldebye: 'ldebye', 'eqperp_ldebye'
        n: 'n'
        eps0: 'eps0', DEFAULTS.PHYSICAL.CONSTANTS_SI['qe']
        q: 'q', 'qe', 'abs_qe', DEFAULTS.PHYSICAL.CONSTANTS_SI['eps0']
        kB: 'kB', DEFAULTS.PHYSICAL.CONSTANTS_SI['kB']
    '''
    provided = dict(ldebye=ldebye, n=n, eps0=eps0, q=q, kB=kB)
    lookup = dict(ldebye=['ldebye', 'eqperp_ldebye'],
                  n='n',
                  eps0='eps0',
                  q=['q', 'qe', 'abs_qe'],
                  kB='kB')
    defaults = dict(eps0=DEFAULTS.PHYSICAL.CONSTANTS_SI['eps0'],
                    q=DEFAULTS.PHYSICAL.CONSTANTS_SI['qe'],
                    kB=DEFAULTS.PHYSICAL.CONSTANTS_SI['kB'])
    vals = xarray_vars_lookup_with_defaults(ds, provided, lookup, defaults)
    return T_from_ldebye(**vals)


''' --------------------- Index in tfbi hypercube --------------------- '''

itAccessor._TFBI_INDEX0_ORDER = ('EBspeed', 'Tgrid', 'ionfrac_H', 'kappae', 'ngrid')
itAccessor.register_attr('tfbi0_order', totype=None,
    value=property(lambda self: self._TFBI_INDEX0_ORDER,
    doc='''standard order of parameters within "type-0" tfbi hypercube, i.e.:
        ('EBspeed', 'Tgrid', 'ionfrac_H', 'kappae', 'ngrid')'''))
itAccessor.register_attr('tfbi0_index_order', totype=None, value=alias('tfbi0_order'))
itAccessor.register_attr('tfbi0keys', totype=None, value=alias('tfbi0_order'))

@itAccessor.register('tfbi0i', aliases=['tfbi0_index'])
def tfbi0_index(array):
    '''returns a 5-tuple of indexes corresponding to this point within "type-0" tfbi hypercube.
    indexes from coords: ('EBspeed_index', 'Tgrid_index', 'ionfrac_H_index', 'kappae_index', 'ngrid_index')
    '''
    coords = [f'{c}_index' for c in array.it.tfbi0_order]
    if not all(c in array.coords for c in coords):
        raise InputError(f'tfbi0_index requires coords {coords}. Only found coords: {list(array.coords)}')
    nonscalars = {c for c in coords if array.coords[c].ndim!=0}
    if not all(array.coords[c].ndim==0 for c in coords):
        raise InputError(f'tfbi0_index requires scalars {coords}. Got non-scalars: {nonscalars}')
    result = tuple(array[c].item() for c in coords)
    if not all(is_integer(i) for i in result):
        raise InputError(f'tfbi0_index requires integer indexes. Got: {result}')
    return tuple(int(i) for i in result)

@itAccessor.register('tfbi0istr', aliases=['tfbi0_index_str'])
def tfbi0_index_str(array):
    '''returns a string of indexes corresponding to this point within "type-0" tfbi hypercube.
    indexes from coords: ('EBspeed_index', 'Tgrid_index', 'ionfrac_H_index', 'kappae_index', 'ngrid_index')
    Format will be like: (0,0,0,0,0)
    '''
    result = tfbi0_index(array)
    return '(' + ",".join(str(i) for i in result) + ')'

@itAccessor.register('tfbi0params', aliases=['tfbi0_params'])
def tfbi0_params(array):
    '''returns a dict of {param: value} for a point within "type-0" tfbi hypercube.
    Result has keys: ('EBspeed', 'Tgrid', 'ionfrac_H', 'kappae', 'ngrid')
    if param values are DataArray scalars, use .item() to convert to scalars.
    '''
    result = {c: array[c] for c in array.it.tfbi0_order}
    result = {c: v.item() if getattr(v, 'ndim', None)==0 else v for c, v in result.items()}
    return result

@itAccessor.register('tfbi0str', aliases=['tfbi0_params_str'])
def tfbi0_params_str(array, fmt='.3g'):
    '''returns a string telling param values for a point within "type-0" tfbi hypercube.
    Result looks like: EBspeed=val, Tgrid=val, ionfrac_H=val, kappae=val, ngrid=val
    '''
    result = tfbi0_params(array)
    return ', '.join(f'{c}={v:{fmt}}' for c, v in result.items())

@itAccessor.register('tfbi0isel', aliases=['tfbi0_index_isel'])
def tfbi0_index_isel(array, tfbii0):
    '''returns a dict of {param: index} given the 5 indexes within "type-0" tfbi hypercube.
    Result has keys: ('EBspeed', 'Tgrid', 'ionfrac_H', 'kappae', 'ngrid'),
    tfbii0: 5-tuple of indexes for those 5 params, in that order ^
    '''
    return dict(zip(array.it.tfbii0_order, tfbii0))
